"""Abstraction for the module 'type' value in a a Runway configuration."""

from __future__ import annotations

import logging
import os
import sys
from typing import TYPE_CHECKING, Any, ClassVar, cast

from typing_extensions import Literal

from ...utils import load_object_from_string

if TYPE_CHECKING:
    from pathlib import Path

    from ...config.models.runway import RunwayModuleTypeTypeDef
    from ...module.base import RunwayModule

LOGGER = logging.getLogger(__name__)


RunwayModuleTypeExtensionsTypeDef = Literal["cdk", "cfn", "k8s", "sls", "tf", "web"]


class RunwayModuleType:
    """Runway configuration ``type`` settings object.

    The ``type`` property of a Runway configuration can be
    used to explicitly specify what module type you are
    intending to deploy.

    Runway determines the type of module you are trying to
    deploy in 3 different ways. First, it will check for the
    ``type`` property as described here, next it will look
    for a suffix as described in :ref:`Module Definition <runway_config:Module>`,
    and finally it will attempt to autodetect your module
    type by scanning the files of the project. If none of
    those settings produces a valid result an error will
    occur. The following are valid explicit types:

    +--------------------+-----------------------------------------------+
    | Type               | IaC Tool/Framework                            |
    +====================+===============================================+
    | ``cdk``            | AWS CDK                                       |
    +--------------------+-----------------------------------------------+
    | ``cloudformation`` | CloudFormation                                |
    +--------------------+-----------------------------------------------+
    | ``serverless``     | Serverless Framework                          |
    +--------------------+-----------------------------------------------+
    | ``terraform``      | Terraform                                     |
    +--------------------+-----------------------------------------------+
    | ``kubernetes``     | Kubernetes                                    |
    +--------------------+-----------------------------------------------+
    | ``static``         | :ref:`index:Static Site`                      |
    +--------------------+-----------------------------------------------+

    Even when specifying a module ``type`` the module structure
    needs to be conducive with that type of project. If the files contained
    within don't match the type then an error will occur.

    """

    EXTENSION_MAP: ClassVar[dict[str, str]] = {
        "cdk": "runway.module.cdk.CloudDevelopmentKit",
        "cfn": "runway.module.cloudformation.CloudFormation",
        "k8s": "runway.module.k8s.K8s",
        "sls": "runway.module.serverless.Serverless",
        "tf": "runway.module.terraform.Terraform",
        "web": "runway.module.staticsite.handler.StaticSite",
    }

    TYPE_MAP: ClassVar[dict[str, str]] = {
        "cdk": EXTENSION_MAP["cdk"],
        "cloudformation": EXTENSION_MAP["cfn"],
        "kubernetes": EXTENSION_MAP["k8s"],
        "serverless": EXTENSION_MAP["sls"],
        "static": EXTENSION_MAP["web"],
        "terraform": EXTENSION_MAP["tf"],
    }

    def __init__(
        self,
        path: Path,
        class_path: str | None = None,
        type_str: RunwayModuleTypeTypeDef | None = None,
    ) -> None:
        """Instantiate class.

        Keyword Args:
            path: The required path to the module
            class_path: A supplied class_path to override the autodetected one.
            type_str: An explicit type to assign to the RunwayModuleType

        """
        self.path = path
        self.class_path = class_path
        self.type_str = type_str
        self.module_class = self._determine_module_class()

    def _determine_module_class(self) -> type[RunwayModule[Any]]:
        """Determine type of module and return deployment module class.

        Returns:
            object: The specified module class

        """
        if self.class_path:
            LOGGER.debug(
                'module class "%s" determined from explicit definition',
                self.class_path,
            )

        if not self.class_path and self.type_str:
            self.class_path = self.TYPE_MAP.get(self.type_str, None)
            if self.class_path:
                LOGGER.debug('module class "%s" determined from explicit type', self.class_path)

        if not self.class_path:
            self._set_class_path_based_on_extension()
            if self.class_path:
                LOGGER.debug('module class "%s" determined from path extension', self.class_path)

        if not self.class_path:
            self._set_class_path_based_on_autodetection()

        if not self.class_path:
            LOGGER.error(
                'module class could not be determined from path "%s"',
                self.path.name,
            )
            sys.exit(1)

        return cast(type["RunwayModule[Any]"], load_object_from_string(self.class_path))

    def _set_class_path_based_on_extension(self) -> None:
        """Based on the directory suffix set the class_path."""
        basename = self.path.name
        basename_split = basename.split(".")
        extension = basename_split[len(basename_split) - 1]
        self.class_path = self.EXTENSION_MAP.get(extension, None)
        if self.class_path:
            LOGGER.debug(
                'module class "%s" determined from path extension "%s"',
                self.class_path,
                extension,
            )

    def _set_class_path_based_on_autodetection(self) -> None:
        """Based on the files detected in the base path set the class_path."""
        if (
            any(
                (self.path / sls).is_file()
                for sls in ["serverless.js", "serverless.ts", "serverless.yml"]
            )
            and (self.path / "package.json").is_file()
        ):
            self.class_path = self.TYPE_MAP.get("serverless", None)
        elif next(self.path.glob("*.tf"), None):
            self.class_path = self.TYPE_MAP.get("terraform", None)
        elif (self.path / "cdk.json").is_file() and (self.path / "package.json").is_file():
            self.class_path = self.TYPE_MAP.get("cdk", None)
        elif (self.path / "overlays").is_dir() and self._find_kustomize_files():
            self.class_path = self.TYPE_MAP.get("kubernetes", None)
        elif (
            next(self.path.glob("*.env"), None)
            or next(self.path.glob("*.yaml"), None)
            or next(self.path.glob("*.yml"), None)
        ):
            self.class_path = self.TYPE_MAP.get("cloudformation", None)
        if self.class_path:
            LOGGER.debug('module class "%s" determined from autodetection', self.class_path)

    def _find_kustomize_files(self) -> bool:
        """Return true if kustomize yaml file found.

        Returns:
            boolean: Whether the kustomize yaml exist

        """
        for _root, _dirnames, filenames in os.walk(self.path):
            for filename in filenames:
                if filename == "kustomization.yaml":
                    return True
        return False



"""Runway config models."""

from __future__ import annotations

import locale
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, TypeVar, cast

import yaml
from packaging.specifiers import InvalidSpecifier, SpecifierSet
from pydantic import (
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic_core import CoreSchema, core_schema

from ....utils.pydantic_validators import LaxStr
from .. import utils
from ..base import ConfigProperty
from ..utils import RUNWAY_LOOKUP_STRING_ERROR, RUNWAY_LOOKUP_STRING_REGEX
from ._assume_role import RunwayAssumeRoleDefinitionModel
from ._builtin_tests import RunwayTestDefinitionModel, ValidRunwayTestTypeValues
from ._future import RunwayFutureDefinitionModel
from ._module import RunwayModuleDefinitionModel
from ._region import RunwayDeploymentRegionDefinitionModel
from ._type_defs import (
    RunwayEnvironmentsType,
    RunwayEnvironmentsUnresolvedType,
    RunwayEnvVarsType,
    RunwayEnvVarsUnresolvedType,
    RunwayModuleTypeTypeDef,
)
from ._variables import RunwayVariablesDefinitionModel

if TYPE_CHECKING:
    from pydantic import BaseModel
    from pydantic.json_schema import JsonSchemaValue
    from typing_extensions import Self

    Model = TypeVar("Model", bound=BaseModel)

LOGGER = logging.getLogger(__name__)


__all__ = [
    "RUNWAY_LOOKUP_STRING_ERROR",
    "RUNWAY_LOOKUP_STRING_REGEX",
    "RunwayAssumeRoleDefinitionModel",
    "RunwayConfigDefinitionModel",
    "RunwayDeploymentDefinitionModel",
    "RunwayDeploymentRegionDefinitionModel",
    "RunwayEnvVarsType",
    "RunwayEnvVarsUnresolvedType",
    "RunwayEnvironmentsType",
    "RunwayEnvironmentsUnresolvedType",
    "RunwayFutureDefinitionModel",
    "RunwayModuleDefinitionModel",
    "RunwayModuleTypeTypeDef",
    "RunwayTestDefinitionModel",
    "RunwayVariablesDefinitionModel",
    "RunwayVersionField",
    "ValidRunwayTestTypeValues",
]


def _deployment_json_schema_extra(schema: dict[str, Any]) -> None:
    """Process the schema after it has been generated.

    Schema is modified in place. Return value is ignored.

    https://pydantic-docs.helpmanual.io/usage/schema/#schema-customization

    """
    schema["description"] = "A collection of modules, regions, and other configurations to deploy."
    # modify schema to allow simple string or mapping definition for a module
    module_ref = schema["properties"]["modules"]["items"].pop("$ref")
    schema["properties"]["modules"]["items"]["anyOf"] = [
        {"$ref": module_ref},
        {"type": "string"},
    ]


class RunwayDeploymentDefinitionModel(ConfigProperty):
    """Model for a Runway deployment definition."""

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra=_deployment_json_schema_extra,
        title="Runway Deployment Definition",
        validate_default=True,
        validate_assignment=True,
    )

    account_alias: Annotated[
        str | None,
        Field(
            description="Used to verify the currently assumed role or credentials. "
            "(supports lookups)",
            examples=["example-alias", "${var alias.${env DEPLOY_ENVIRONMENT}}"],
        ),
    ] = None
    """Used to verify the currently assumed role or credentials. (supports lookups)"""

    account_id: Annotated[
        str | None,
        LaxStr,
        Field(
            description="Used to verify the currently assumed role or credentials. "
            "(supports lookups)",
            examples=["123456789012", "${var id.${env DEPLOY_ENVIRONMENT}}"],
        ),
    ] = None
    """Used to verify the currently assumed role or credentials. (supports lookups)"""

    assume_role: Annotated[
        str | RunwayAssumeRoleDefinitionModel | None,
        Field(
            description="Assume a role when processing the deployment. (supports lookups)",
            examples=[
                "arn:aws:iam::123456789012:role/name",
                *cast(
                    "dict[str, list[str]]",
                    RunwayAssumeRoleDefinitionModel.model_config.get("json_schema_extra", {}),
                ).get("examples", []),
            ],
        ),
    ] = RunwayAssumeRoleDefinitionModel()
    """Assume a role when processing the deployment. (supports lookups)"""

    env_vars: Annotated[
        RunwayEnvVarsUnresolvedType,
        Field(
            title="Environment Variables",
            description="Additional variables to add to the environment when "
            "processing the deployment. (supports lookups)",
            examples=[
                "${var env_vars.${env DEPLOY_ENVIRONMENT}}",
                {
                    "EXAMPLE_VARIABLE": "value",
                    "KUBECONFIG": [".kube", "${env DEPLOY_ENVIRONMENT}", "config"],
                },
            ],
        ),
    ] = {}
    """Additional variables to add to the environment when processing the deployment. (supports lookups)"""

    environments: Annotated[
        RunwayEnvironmentsUnresolvedType,
        Field(
            description="Explicitly enable/disable the deployment for a specific "
            "deploy environment, AWS Account ID, and AWS Region combination. "
            "Can also be set as a static boolean value. (supports lookups)",
            examples=[
                "${var envs.${env DEPLOY_ENVIRONMENT}}",
                {"dev": "123456789012", "prod": "us-east-1"},
                {"dev": True, "prod": False},
                {"dev": ["us-east-1"], "prod": ["us-west-2", "ca-central-1"]},
                {
                    "dev": ["123456789012/us-east-1", "123456789012/us-west-2"],
                    "prod": ["234567890123/us-east-1", "234567890123/us-west-2"],
                },
            ],
        ),
    ] = {}
    """Explicitly enable/disable the deployment for a specific deploy environment,
    AWS Account ID, and AWS Region combination.
    Can also be set as a static boolean value. (supports lookups)

    """

    modules: Annotated[
        list[RunwayModuleDefinitionModel],
        Field(description="An array of modules to process as part of a deployment."),
    ]
    """An array of modules to process as part of a deployment."""

    module_options: Annotated[
        dict[str, Any] | str,
        Field(
            description="Options that are passed directly to the modules within this deployment. "
            "(supports lookups)",
            examples=[
                "${var sampleapp.options.${env DEPLOY_ENVIRONMENT}}",
                {"some_option": "value"},
            ],
        ),
    ] = {}
    """Options that are passed directly to the modules within this deployment. (supports lookups)"""

    name: Annotated[
        str,
        Field(
            description="The name of the deployment to be displayed in logs and the "
            "interactive selection menu.",
        ),
    ] = "unnamed_deployment"
    """The name of the deployment to be displayed in logs and the interactive selection menu."""

    parallel_regions: Annotated[
        list[str] | str,
        Field(
            description="An array of AWS Regions to process asynchronously. (supports lookups)",
            examples=[
                ["us-east-1", "us-west-2"],
                "${var regions.${dev DEPLOY_ENVIRONMENT}}",
            ],
        ),
    ] = []
    """An array of AWS Regions to process asynchronously. (supports lookups)"""

    parameters: Annotated[
        dict[str, Any] | str,
        Field(
            description="Used to pass variable values to modules in place of an "
            "environment configuration file. (supports lookups)",
            examples=[
                {"namespace": "example-${env DEPLOY_ENVIRONMENT}"},
                "${var sampleapp.parameters.${env DEPLOY_ENVIRONMENT}}",
            ],
        ),
    ] = {}
    """Used to pass variable values to modules in place of an environment configuration file. (supports lookups)"""

    regions: Annotated[
        list[str] | str,
        Field(
            description="An array of AWS Regions to process this deployment in. (supports lookups)",
            examples=[
                ["us-east-1", "us-west-2"],
                "${var regions.${dev DEPLOY_ENVIRONMENT}}",
                *cast(
                    "dict[str, list[str]]",
                    RunwayDeploymentRegionDefinitionModel.model_config.get("json_schema_extra", {}),
                ).get("examples", []),
            ],
        ),
    ] = []
    """An array of AWS Regions to process this deployment in. (supports lookups)"""

    @model_validator(mode="before")
    @classmethod
    def _convert_simple_module(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Convert simple modules to dicts."""
        modules = values.get("modules", [])
        result: list[dict[str, Any]] = []
        for module in modules:
            if isinstance(module, str):
                result.append({"path": module})
            else:
                result.append(module)
        values["modules"] = result
        return values

    @model_validator(mode="before")
    @classmethod
    def _validate_regions(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate & simplify regions."""
        raw_regions: str | list[str] = values.get("regions", [])
        parallel_regions = values.get("parallel_regions", [])
        if all(isinstance(i, str) for i in [raw_regions, parallel_regions]):
            raise ValueError(
                "unable to validate parallel_regions/regions - both are defined as strings"
            )
        if any(isinstance(i, str) for i in [raw_regions, parallel_regions]):
            return values  # one is a lookup so skip the remainder of the checks
        if isinstance(raw_regions, list):
            regions = raw_regions
        else:
            regions = RunwayDeploymentRegionDefinitionModel.model_validate(raw_regions)

        if regions and parallel_regions:
            raise ValueError("only one of parallel_regions or regions can be defined")
        if not regions and not parallel_regions:
            raise ValueError("either parallel_regions or regions must be defined")

        if isinstance(regions, RunwayDeploymentRegionDefinitionModel):
            values["regions"] = []
            values["parallel_regions"] = regions.parallel
        return values

    _validate_string_is_lookup = field_validator(
        "env_vars",
        "environments",
        "module_options",
        "parallel_regions",
        "parameters",
        "regions",
        mode="before",
    )(utils.validate_string_is_lookup)


class RunwayVersionField(SpecifierSet):
    """Extends packaging.specifiers.SpecifierSet for use with pydantic."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Yield one of more validators with will be called to validate the input.

        Each validator will receive, as input, the value returned from the previous validator.

        """
        assert source_type is RunwayVersionField  # noqa: S101
        return core_schema.no_info_before_validator_function(
            cls._convert_value,
            core_schema.any_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                str, info_arg=False, return_schema=core_schema.str_schema()
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.JsonSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:  # cov: ignore
        """Mutate the field schema in place.

        This is only called when output JSON schema from a model.

        """
        json_schema = handler(schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema["type"] = "str"
        return json_schema

    @classmethod
    def _convert_value(cls, v: str | SpecifierSet) -> RunwayVersionField:
        """Convert runway_version string into SpecifierSet with some value handling.

        Args:
            v: The value to be converted/validated.

        Raises:
            ValueError: The provided value is not a valid version specifier set.

        """
        if isinstance(v, (cls, SpecifierSet)):
            return RunwayVersionField(str(v), prereleases=True)
        try:
            return RunwayVersionField(v, prereleases=True)
        except InvalidSpecifier:
            if any(v.startswith(i) for i in ["!", "~", "<", ">", "="]):
                raise ValueError(f"{v} is not a valid version specifier set") from None
            LOGGER.debug(
                "runway_version is not a valid version specifier; trying as an exact version",
                exc_info=True,
            )
            return RunwayVersionField("==" + v, prereleases=True)


class RunwayConfigDefinitionModel(ConfigProperty):
    """Runway configuration definition model."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        json_schema_extra={
            "description": "Configuration file for use with Runway.",
        },
        extra="forbid",
        title="Runway Configuration File",
        validate_default=True,
        validate_assignment=True,
    )

    deployments: Annotated[
        list[RunwayDeploymentDefinitionModel],
        Field(description="Array of Runway deployments definitions."),
    ] = []
    """Array of Runway deployments definitions."""

    future: Annotated[
        RunwayFutureDefinitionModel,
        Field(description="Enable future features before they become default behavior."),
    ] = RunwayFutureDefinitionModel()
    """Enable future features before they become default behavior."""

    ignore_git_branch: Annotated[
        bool,
        Field(
            description="Optionally exclude the git branch name when determining the "
            "current deploy environment.",
        ),
    ] = False
    """Optionally exclude the git branch name when determining the current deploy environment."""

    runway_version: Annotated[
        RunwayVersionField | None,
        Field(
            description="Define the versions of Runway that can be used with this "
            "configuration file.",
            examples=['"<2.0.0"', '"==1.14.0"', '">=1.14.0,<2.0.0"'],
        ),
    ] = None
    """Define the versions of Runway that can be used with this configuration file."""

    tests: Annotated[
        list[RunwayTestDefinitionModel],
        Field(
            description="Array of Runway test definitions that are executed with the 'test' command."
        ),
    ] = []

    variables: Annotated[RunwayVariablesDefinitionModel, Field(description="Runway variables.")] = (
        RunwayVariablesDefinitionModel()
    )
    """Runway variables."""

    @model_validator(mode="before")
    @classmethod
    def _add_deployment_names(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Add names to deployments that are missing them."""
        deployments = values.get("deployments", [])
        for i, deployment in enumerate(deployments):
            if not deployment.get("name"):
                deployment["name"] = f"deployment_{i + 1}"
        values["deployments"] = deployments
        return values

    @field_serializer("runway_version", when_used="json-unless-none")
    def _serialize_runway_version(self, runway_version: RunwayVersionField, _info: Any) -> str:
        """Serialize ``runway_version`` field when dumping to JSON."""
        return str(runway_version)

    @classmethod
    def parse_file(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls: type[Self], path: str | Path
    ) -> Self:
        """Parse a file."""
        return cls.model_validate(
            yaml.safe_load(
                Path(path).read_text(encoding=locale.getpreferredencoding(do_setlocale=False))
            )
        )



"""Hook data models.

These are makeshift data models for use until Runway v2 is released and pydantic
can be used.

"""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any, ClassVar, cast

from docker.models.images import Image
from pydantic import ConfigDict, Field, PrivateAttr, model_validator

from ....core.providers.aws import AccountDetails
from ....utils import BaseModel, MutableMap

if TYPE_CHECKING:
    from ....context import CfnginContext

ECR_REPO_FQN_TEMPLATE = "{aws_account_id}.dkr.ecr.{aws_region}.amazonaws.com/{repo_name}"


class ElasticContainerRegistry(BaseModel):
    """AWS Elastic Container Registry."""

    PUBLIC_URI_TEMPLATE: ClassVar[str] = "public.ecr.aws/{registry_alias}/"
    URI_TEMPLATE: ClassVar[str] = "{aws_account_id}.dkr.ecr.{aws_region}.amazonaws.com/"

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    account_id: str | None = None
    """AWS account ID that owns the registry being logged into."""

    alias: str | None = None
    """If it is a public repository, the alias of the repository."""

    public: bool = True
    """Whether the repository is public."""

    region: str | None = Field(default=None, alias="aws_region")
    """AWS region where the registry is located."""

    @property
    def fqn(self) -> str:
        """Fully qualified ECR name."""
        if self.public:
            return self.PUBLIC_URI_TEMPLATE.format(registry_alias=self.alias)
        return self.URI_TEMPLATE.format(aws_account_id=self.account_id, aws_region=self.region)

    @model_validator(mode="before")
    @classmethod
    def _set_defaults(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Set default values based on other values."""
        values.setdefault("public", bool(values.get("alias")))

        if not values["public"]:
            account_id = values.get("account_id")
            ctx: CfnginContext | None = values.get("context")
            aws_region = values.get("aws_region")
            if not ctx and not (account_id or aws_region):
                raise ValueError("context is required to resolve values")
            if ctx:
                if not account_id:
                    values["account_id"] = AccountDetails(ctx).id
                if not aws_region:
                    values["aws_region"] = ctx.env.aws_region or "us-east-1"
        return values


class DockerImage(BaseModel):
    """Wrapper for :class:`docker.models.images.Image`."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    _repo: str | None = PrivateAttr(default=None)
    image: Image

    @property
    def id(self) -> str:
        """ID of the image."""
        return self.image.id

    @property
    def repo(self) -> str:
        """Repository URI."""
        if not self._repo:
            self._repo = self.image.attrs["RepoTags"][0].rsplit(":", 1)[0]
        return cast(str, self._repo)

    @property
    def short_id(self) -> str:
        """ID of the image truncated to 10 characters plus the ``sha256:`` prefix."""
        return self.image.short_id

    @property
    def tags(self) -> list[str]:
        """List of image tags."""
        self.image.reload()
        return [uri.split(":")[-1] for uri in self.image.tags]

    @property
    def uri(self) -> MutableMap:
        """Return a mapping of tag to image URI."""
        return MutableMap(**{uri.split(":")[-1]: uri for uri in self.image.tags})

    def __bool__(self) -> bool:
        """Evaluate the boolean value of the object instance."""
        return True


class ElasticContainerRegistryRepository(BaseModel):
    """AWS Elastic Container Registry (ECR) Repository."""

    model_config = ConfigDict(populate_by_name=True)

    name: Annotated[str, Field(alias="repo_name")]
    """The name of the repository."""

    registry: ElasticContainerRegistry
    """Information about an ECR registry."""

    @property
    def fqn(self) -> str:
        """Fully qualified ECR repo name."""
        return self.registry.fqn + self.name



"""Runway static site Module options."""

from __future__ import annotations

from pathlib import Path
from typing import Any, cast

from pydantic import ConfigDict, model_validator

from ....config.models.base import ConfigProperty


class RunwayStaticSiteExtraFileDataModel(ConfigProperty):
    """Model for Runway static site Module extra_files option item."""

    model_config = ConfigDict(
        extra="forbid",
        title="Runway static site Module extra_files option item",
        validate_assignment=True,
        validate_default=True,
    )

    content_type: str | None = None
    """An explicit content type for the file.
    If not provided, will attempt to determine based on the name provided.

    """

    content: Any = None
    """Inline content that will be used as the file content.
    This or ``file`` must be provided.

    """

    file: Path | None = None
    """Path to an existing file.
    The content of this file will be uploaded to the static site S3 bucket using
    the name as the object key.
    This or ``content`` must be provided.

    """

    name: str
    """The destination name of the file to create."""

    @model_validator(mode="before")
    @classmethod
    def _autofill_content_type(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Attempt to fill content_type if not provided."""
        if values.get("content_type"):
            return values
        name = cast(str, values.get("name", ""))
        if name.endswith(".json"):
            values["content_type"] = "application/json"
        elif name.endswith((".yaml", ".yml")):
            values["content_type"] = "text/yaml"
        return values

    @model_validator(mode="before")
    @classmethod
    def _validate_content_or_file(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate that content or file is provided."""
        if all(i in values and values[i] for i in ["content", "file"]):
            raise ValueError("only one of content or file can be provided")
        if not any(i in values for i in ["content", "file"]):
            raise ValueError("one of content or file must be provided")
        return values


class RunwayStaticSitePreBuildStepDataModel(ConfigProperty):
    """Model for Runway static site Module pre_build_steps option item."""

    model_config = ConfigDict(
        extra="forbid",
        title="Runway static site Module pre_build_steps option item",
        validate_default=True,
        validate_assignment=True,
    )

    command: str
    """The command to run."""

    cwd: Path = Path.cwd()
    """The working directory for the subprocess running the command.
    If not provided, the current working directory is used.

    """


class RunwayStaticSiteSourceHashingDirectoryDataModel(ConfigProperty):
    """Model for Runway static site Module source_hashing.directory option item."""

    model_config = ConfigDict(
        extra="forbid",
        title="Runway static site Module source_hashing.directories option item",
        validate_default=True,
        validate_assignment=True,
    )

    exclusions: list[str] = []
    """List of gitignore formmated globs to ignore when calculating the hash."""

    path: Path
    """Path to files to include in the hash."""


class RunwayStaticSiteSourceHashingDataModel(ConfigProperty):
    """Model for Runway static site Module source_hashing option."""

    model_config = ConfigDict(
        extra="forbid",
        title="Runway static site Module source_hashing option",
        validate_default=True,
        validate_assignment=True,
    )

    directories: list[RunwayStaticSiteSourceHashingDirectoryDataModel] = [
        RunwayStaticSiteSourceHashingDirectoryDataModel(path="./")  # type: ignore
    ]
    """Explicitly provide the directories to use when calculating the hash.
    If not provided, will default to the root of the module.
    """

    enabled: bool = True
    """Enable source hashing. If not enabled, build and upload will occur on every deploy."""

    parameter: str | None = None
    """SSM parameter where the hash of each build is stored."""


class RunwayStaticSiteModuleOptionsDataModel(ConfigProperty):
    """Model for Runway static site Module options."""

    model_config = ConfigDict(
        extra="ignore",
        title="Runway static site Module options",
        validate_default=True,
        validate_assignment=True,
    )

    build_output: str = "./"
    """Directory where build output is placed. Defaults to current working directory."""

    build_steps: list[str] = []
    """List of commands to run to build the static site."""

    extra_files: list[RunwayStaticSiteExtraFileDataModel] = []
    """List of files that should be uploaded to S3 after the build.
    Used to dynamically create or select file.
    """

    pre_build_steps: list[RunwayStaticSitePreBuildStepDataModel] = []
    """Commands to be run prior to the build process."""

    source_hashing: RunwayStaticSiteSourceHashingDataModel = (
        RunwayStaticSiteSourceHashingDataModel()
    )
    """Overrides for source hash calculation and tracking."""



"""Utilities for gen-sample commands."""

import logging
import shutil
from pathlib import Path

import click
from cfn_flip import to_yaml  # pyright: ignore[reportUnknownVariableType]
from typing_extensions import Literal

from ....blueprints.tf_state import TfState
from ....context import CfnginContext

LOGGER = logging.getLogger(__name__.replace("._", "."))
ROOT = Path(__file__).parent.parent.parent.parent
TEMPLATES = ROOT / "templates"


def convert_gitignore(src: Path) -> Path:
    """Rename a gitignore template.

    Keyword Args:
        src: Path object for source file.

    Returns:
        The renamed file if it was created.

    """
    gitignore = src.parent / ".gitignore"
    LOGGER.debug('renaming "%s" to "%s"', src, gitignore)
    src.rename(gitignore)
    return gitignore


def copy_sample(ctx: click.Context, src: Path, dest: Path) -> None:
    """Copy a sample directory.

    Args:
        ctx: Click context object.
        src: Source path.
        dest: Destination path.

    """
    if dest.exists():
        LOGGER.error("Directory %s already exists!", dest)
        ctx.exit(1)
    LOGGER.debug('copying "%s" to "%s"', src, dest)
    shutil.copytree(src, dest)


def write_tfstate_template(
    dest: Path, *, bucket_deletion_policy: Literal["Delete", "Retain"] = "Retain"
) -> None:
    """Write TfState blueprint as a YAML CFN template.

    Args:
        dest: File to be written to.
        bucket_deletion_policy: CloudFormation deletion policy for S3 Bucket.

    """
    LOGGER.debug('writing TfState as a YAML template to "%s"', dest)
    dest.write_text(
        to_yaml(
            TfState("test", CfnginContext(environment={"namespace": "test"})).to_json(
                {"BucketDeletionPolicy": bucket_deletion_policy}
            )
        )
    )



"""Pipenv interface."""

from __future__ import annotations

import locale
import logging
import re
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

from ..compat import cached_property
from ..exceptions import RunwayError
from ..utils import Version
from .base_classes import DependencyManager

if TYPE_CHECKING:
    from _typeshed import StrPath

LOGGER = logging.getLogger(__name__)


class PipenvExportFailedError(RunwayError):
    """Pipenv export failed to produce a ``requirements.txt`` file."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Instantiate class. All args/kwargs are passed to parent method."""
        self.message = (
            "pipenv lock to requirements.txt format failed; review pipenv's"
            " output above to troubleshoot"
        )
        super().__init__(*args, **kwargs)


class PipenvNotFoundError(RunwayError):
    """Pipenv not installed or found in $PATH."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Instantiate class. All args/kwargs are passed to parent method."""
        self.message = (
            "pipenv not installed or not in PATH! "
            "Install it according to pipenv docs (https://pipenv.pypa.io/en/latest/) "
            "and ensure it is available in PATH."
        )
        super().__init__(*args, **kwargs)


class Pipenv(DependencyManager):
    """Pipenv dependency manager."""

    CONFIG_FILES: ClassVar[tuple[str, ...]] = (
        "Pipfile",
        "Pipfile.lock",
    )
    """Configuration files used by pipenv."""

    EXECUTABLE: ClassVar[str] = "pipenv"
    """CLI executable."""

    @cached_property
    def version(self) -> Version:
        """pipenv version."""
        cmd_output = self._run_command([self.EXECUTABLE, "--version"])
        match = re.search(r"^pipenv, version (?P<version>\S*)", cmd_output)
        if not match:
            LOGGER.warning("unable to parse pipenv version from output:\n%s", cmd_output)
            return Version("0.0.0")
        return Version(match.group("version"))

    @classmethod
    def dir_is_project(cls, directory: StrPath, **__kwargs: Any) -> bool:
        """Determine if the directory contains a project for this dependency manager.

        Args:
            directory: Directory to check.

        """
        dir_path = Path(directory)
        if not (dir_path / Pipenv.CONFIG_FILES[0]).is_file():
            return False

        if not (dir_path / Pipenv.CONFIG_FILES[1]).is_file():
            LOGGER.warning("%s not found", Pipenv.CONFIG_FILES[1])
        return True

    def export(self, *, dev: bool = False, output: StrPath) -> Path:
        """Export the lock file to other formats (requirements.txt only).

        The underlying command being executed by this method is
        ``pipenv lock --requirements``.

        Args:
            dev: Include development dependencies.
            output: Path to the output file.

        """
        output = Path(output)
        try:
            result = self._run_command(
                self.generate_command(
                    "lock",
                    dev=dev,
                    requirements=True,
                ),
                suppress_output=True,
            )
        except subprocess.CalledProcessError as exc:
            raise PipenvExportFailedError from exc
        output.parent.mkdir(exist_ok=True, parents=True)  # ensure directory exists
        output.write_text(str(result), encoding=locale.getpreferredencoding(do_setlocale=False))
        return output



"""Dedicated lookup for use with :class:`~runway.cfngin.hooks.awslambda.base_classes.AwsLambdaHook` based hooks.

To use this hook, there must be a
:class:`~runway.cfngin.hooks.awslambda.base_classes.AwsLambdaHook` based hook defined
in the :attr:`~cfngin.config.pre_deploy` section of the CFNgin configuration file.
This hook must also define a :attr:`~cfngin.hook.data_key` that is unique within
the CFNgin configuration file (it can be reused in other CFNgin configuration files).
The :attr:`~cfngin.hook.data_key` is then passed to the lookup as it's input/query.
This allows the lookup to function during a ``runway plan``.

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar, cast

from pydantic import ValidationError
from troposphere.awslambda import Code, Content

from ....lookups.handlers.base import LookupHandler
from ....utils import load_object_from_string

if TYPE_CHECKING:

    from ....config import CfnginConfig
    from ....config.models.cfngin import CfnginHookDefinitionModel
    from ....context import CfnginContext
    from ...hooks.awslambda.base_classes import AwsLambdaHook
    from ...hooks.awslambda.models.responses import AwsLambdaHookDeployResponse

LOGGER = logging.getLogger(__name__)


class AwsLambdaLookup(LookupHandler["CfnginContext"]):
    """Lookup for AwsLambdaHook responses."""

    TYPE_NAME: ClassVar[str] = "awslambda"

    @classmethod
    def get_deployment_package_data(
        cls, context: CfnginContext, data_key: str
    ) -> AwsLambdaHookDeployResponse:
        """Get the response of an AwsLambdaHook run.

        Args:
            context: CFNgin context object.
            data_key: The value of the ``data_key`` field as assigned in a
                Hook definition.

        Returns:
            The :class:`~runway.cfngin.hooks.awslambda.base_classes.AwsLambdaHook`
            response parsed into a data model.
            This will come from hook data if it exists or it will be calculated
            and added to hook data for future use.

        Raises:
            TypeError: The data stored in hook data does not align with the
                expected data model.

        """
        # needs to be imported here to avoid cyclic imports for conditional code
        # caused by import of runway.cfngin.actions.deploy in runway.cfngin.hooks.base
        from ...hooks.awslambda.models.responses import (
            AwsLambdaHookDeployResponse as _AwsLambdaHookDeployResponse,
        )

        if data_key not in context.hook_data:
            LOGGER.debug("%s missing from hook_data; attempting to get value", data_key)
            hook = cls.init_hook_class(
                context, cls.get_required_hook_definition(context.config, data_key)
            )
            context.set_hook_data(data_key, hook.plan())
        try:
            return _AwsLambdaHookDeployResponse.model_validate(context.hook_data[data_key])
        except ValidationError:
            raise TypeError(
                "expected AwsLambdaHookDeployResponseTypedDict, "
                f"not {context.hook_data[data_key]}"
            ) from None

    @staticmethod
    def get_required_hook_definition(
        config: CfnginConfig, data_key: str
    ) -> CfnginHookDefinitionModel:
        """Get the required Hook definition from the CFNgin config.

        Currently, this only supports finding the data_key pre_deploy.

        Args:
            config: CFNgin config being processed.
            data_key: The value of the ``data_key`` field as assigned in a
                Hook definition.

        Returns:
            The Hook definition set to use the provided ``data_key``.

        Raises:
            ValueError: Either a Hook definition was not found for the provided
                ``data_key`` or, more than one was found.

        """
        hooks_with_data_key = [
            hook_def for hook_def in config.pre_deploy if hook_def.data_key == data_key
        ]
        if not hooks_with_data_key:
            raise ValueError(f"no hook definition found with data_key {data_key}")
        if len(hooks_with_data_key) > 1:
            raise ValueError(f"more than one hook definition found with data_key {data_key}")
        return hooks_with_data_key.pop()

    @classmethod
    def handle(
        cls, value: str, context: CfnginContext, **_kwargs: Any
    ) -> AwsLambdaHookDeployResponse:
        """Retrieve metadata for an AWS Lambda deployment package.

        Args:
            value: Value to resolve.
            context: The current context object.

        Returns:
            The full :class:`~awslambda.models.response.AwsLambdaHookDeployResponse`
            data model.

        """
        query, _ = cls.parse(value)
        return cls.get_deployment_package_data(context, query)

    @staticmethod
    def init_hook_class(
        context: CfnginContext, hook_def: CfnginHookDefinitionModel
    ) -> AwsLambdaHook[Any]:
        """Initialize AwsLambdaHook subclass instance.

        Args:
            context: CFNgin context object.
            hook_def: The :class:`~runway.cfngin.hooks.awslambda.base_classes.AwsLambdaHook`
                definition.

        Returns:
            The loaded AwsLambdaHook object.

        """
        # needs to be imported here to avoid cyclic imports for conditional code
        # caused by import of runway.cfngin.actions.deploy in runway.cfngin.hooks.base
        from ...hooks.awslambda.base_classes import AwsLambdaHook as _AwsLambdaHook

        kls = load_object_from_string(hook_def.path)
        if (
            not isinstance(kls, type)
            or not hasattr(kls, "__subclasscheck__")
            or not issubclass(kls, _AwsLambdaHook)
        ):
            raise TypeError(
                f"hook path {hook_def.path} for hook with data_key {hook_def.data_key} "
                "must be a subclass of AwsLambdaHook to use this lookup"
            )
        return cast("AwsLambdaHook[Any]", kls(context, **hook_def.args))

    class Code(LookupHandler["CfnginContext"]):
        """Lookup for AwsLambdaHook responses."""

        TYPE_NAME: ClassVar[str] = "awslambda.Code"

        @classmethod
        def handle(cls, value: str, context: CfnginContext, *args: Any, **kwargs: Any) -> Code:
            """Retrieve metadata for an AWS Lambda deployment package.

            Args:
                value: Value to resolve.
                context: The current context object.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Value that can be passed into CloudFormation property
                ``AWS::Lambda::Function.Code``.

            """
            return Code(
                **AwsLambdaLookup.handle(value, context, *args, **kwargs).model_dump(
                    by_alias=True,
                    exclude_none=True,
                    include={"bucket_name", "object_key", "object_version_id"},
                )
            )

    class CodeSha256(LookupHandler["CfnginContext"]):
        """Lookup for AwsLambdaHook responses."""

        TYPE_NAME: ClassVar[str] = "awslambda.CodeSha256"

        @classmethod
        def handle(cls, value: str, context: CfnginContext, *args: Any, **kwargs: Any) -> str:
            """Retrieve metadata for an AWS Lambda deployment package.

            Args:
                value: Value to resolve.
                context: The current context object.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Value that can be passed into CloudFormation property
                ``AWS::Lambda::Version.CodeSha256``.

            """
            return AwsLambdaLookup.handle(value, context, *args, **kwargs).code_sha256

    class CompatibleArchitectures(LookupHandler["CfnginContext"]):
        """Lookup for AwsLambdaHook responses."""

        TYPE_NAME: ClassVar[str] = "awslambda.CompatibleArchitectures"

        @classmethod
        def handle(
            cls, value: str, context: CfnginContext, *args: Any, **kwargs: Any
        ) -> list[str] | None:
            """Retrieve metadata for an AWS Lambda deployment package.

            Args:
                value: Value to resolve.
                context: The current context object.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Value that can be passed into CloudFormation property
                ``AWS::Lambda::LayerVersion.CompatibleArchitectures``.

            """
            _query, lookup_args = cls.parse(value)
            return cls.format_results(
                AwsLambdaLookup.handle(value, context, *args, **kwargs).compatible_architectures,
                **lookup_args,
            )

    class CompatibleRuntimes(LookupHandler["CfnginContext"]):
        """Lookup for AwsLambdaHook responses."""

        TYPE_NAME: ClassVar[str] = "awslambda.CompatibleRuntimes"

        @classmethod
        def handle(cls, value: str, context: CfnginContext, *args: Any, **kwargs: Any) -> Any:
            """Retrieve metadata for an AWS Lambda deployment package.

            Args:
                value: Value to resolve.
                context: The current context object.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Value that can be passed into CloudFormation property
                ``AWS::Lambda::LayerVersion.CompatibleRuntimes``.

            """
            _query, lookup_args = cls.parse(value)
            return cls.format_results(
                AwsLambdaLookup.handle(value, context, *args, **kwargs).compatible_runtimes,
                **lookup_args,
            )

    class Content(LookupHandler["CfnginContext"]):
        """Lookup for AwsLambdaHook responses."""

        TYPE_NAME: ClassVar[str] = "awslambda.Content"

        @classmethod
        def handle(cls, value: str, context: CfnginContext, *args: Any, **kwargs: Any) -> Content:
            """Retrieve metadata for an AWS Lambda deployment package.

            Args:
                value: Value to resolve.
                context: The current context object.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Value that can be passed into CloudFormation property
                ``AWS::Lambda::LayerVersion.Content``.

            """
            return Content(
                **AwsLambdaLookup.handle(value, context, *args, **kwargs).model_dump(
                    by_alias=True,
                    exclude_none=True,
                    include={"bucket_name", "object_key", "object_version_id"},
                )
            )

    class LicenseInfo(LookupHandler["CfnginContext"]):
        """Lookup for AwsLambdaHook responses."""

        TYPE_NAME: ClassVar[str] = "awslambda.LicenseInfo"

        @classmethod
        def handle(
            cls, value: str, context: CfnginContext, *args: Any, **kwargs: Any
        ) -> str | None:
            """Retrieve metadata for an AWS Lambda deployment package.

            Args:
                value: Value to resolve.
                context: The current context object.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Value that can be passed into CloudFormation property
                ``AWS::Lambda::LayerVersion.LicenseInfo``.

            """
            _query, lookup_args = cls.parse(value)
            return cls.format_results(
                AwsLambdaLookup.handle(value, context, *args, **kwargs).license,
                **lookup_args,
            )

    class Runtime(LookupHandler["CfnginContext"]):
        """Lookup for AwsLambdaHook responses."""

        TYPE_NAME: ClassVar[str] = "awslambda.Runtime"

        @classmethod
        def handle(cls, value: str, context: CfnginContext, *args: Any, **kwargs: Any) -> str:
            """Retrieve metadata for an AWS Lambda deployment package.

            Args:
                value: Value to resolve.
                context: The current context object.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Value that can be passed into CloudFormation property
                ``AWS::Lambda::Function.Runtime``.

            """
            return AwsLambdaLookup.handle(value, context, *args, **kwargs).runtime

    class S3Bucket(LookupHandler["CfnginContext"]):
        """Lookup for AwsLambdaHook responses."""

        TYPE_NAME: ClassVar[str] = "awslambda.S3Bucket"

        @classmethod
        def handle(cls, value: str, context: CfnginContext, *args: Any, **kwargs: Any) -> str:
            """Retrieve metadata for an AWS Lambda deployment package.

            Args:
                value: Value to resolve.
                context: The current context object.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Value that can be passed into CloudFormation property
                ``AWS::Lambda::Function.Code.S3Bucket`` or
                ``AWS::Lambda::LayerVersion.Content.S3Bucket``.

            """
            return AwsLambdaLookup.handle(value, context, *args, **kwargs).bucket_name

    class S3Key(LookupHandler["CfnginContext"]):
        """Lookup for AwsLambdaHook responses."""

        TYPE_NAME: ClassVar[str] = "awslambda.S3Key"

        @classmethod
        def handle(cls, value: str, context: CfnginContext, *args: Any, **kwargs: Any) -> str:
            """Retrieve metadata for an AWS Lambda deployment package.

            Args:
                value: Value to resolve.
                context: The current context object.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Value that can be passed into CloudFormation property
                ``AWS::Lambda::Function.Code.S3Key`` or
                ``AWS::Lambda::LayerVersion.Content.S3Key``.

            """
            return AwsLambdaLookup.handle(value, context, *args, **kwargs).object_key

    class S3ObjectVersion(LookupHandler["CfnginContext"]):
        """Lookup for AwsLambdaHook responses."""

        TYPE_NAME: ClassVar[str] = "awslambda.S3ObjectVersion"

        @classmethod
        def handle(
            cls, value: str, context: CfnginContext, *args: Any, **kwargs: Any
        ) -> str | None:
            """Retrieve metadata for an AWS Lambda deployment package.

            Args:
                value: Value to resolve.
                context: The current context object.
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.

            Returns:
                Value that can be passed into CloudFormation property
                ``AWS::Lambda::Function.Code.S3ObjectVersion`` or
                ``AWS::Lambda::LayerVersion.Content.S3ObjectVersion``.

            """
            return AwsLambdaLookup.handle(value, context, *args, **kwargs).object_version_id



"""Comparator.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/comparator.py

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from collections.abc import Generator, Iterator

    from .file_generator import FileStats
    from .sync_strategy.base import BaseSync


LOGGER = logging.getLogger(__name__.replace("._", "."))


class Comparator:
    """Performs all of the comparisons behind the sync operation."""

    def __init__(
        self,
        file_at_src_and_dest_sync_strategy: BaseSync,
        file_not_at_dest_sync_strategy: BaseSync,
        file_not_at_src_sync_strategy: BaseSync,
    ) -> None:
        """Instantiate class."""
        self._sync_strategy = file_at_src_and_dest_sync_strategy
        self._not_at_dest_sync_strategy = file_not_at_dest_sync_strategy
        self._not_at_src_sync_strategy = file_not_at_src_sync_strategy

    def call(  # noqa: C901, PLR0912, PLR0915
        self, src_files: Iterator[FileStats], dest_files: Iterator[FileStats]
    ) -> Generator[FileStats, None, None]:
        """Preform the actual comparisons.

        The parameters this takes are the generated files for both the source
        and the destination. The key concept in this function is that no matter
        the type of where the files are coming from, they are listed in the same
        order, least to greatest in collation order. This allows for easy
        comparisons to determine if file needs to be added or deleted. Comparison
        keys are used to determine if two files are the same and each file has a
        unique comparison key. If they are the same compare the size and last
        modified times to see if a file needs to be updated. Ultimately, it will
        yield a sequence of file info objects that will be sent to the
        ``S3Handler``.

        Algorithm:
            Try to take next from both files. If it is empty signal
            corresponding done flag. If both generated lists are not done
            compare compare_keys. If equal, compare size and time to see if
            it needs to be updated. If source compare_key is less than dest
            compare_key, the file needs to be added to the destination. Take
            the next source file but not not destination file. If the source
            compare_key is greater than dest compare_key, that destination file
            needs to be deleted from the destination. Take the next dest file
            but not the source file. If the source list is empty delete the
            rest of the files in the dest list from the destination. If the
            dest list is empty add the rest of the file in source list to
            the destination.

        Args:
            src_files: The generated FileInfo objects from the source.
            dest_files: The generated FileInfo objects from the dest.

        Returns:
            Yields the FilInfo objects of the files that need to be operated on.

        """
        dest_file: FileStats | None = None
        dest_done = False  # True if there are no more files form the dest left.
        dest_take = True  # Take the next dest file from the generated files if true
        src_file: FileStats | None = None
        src_done = False  # True if there are no more files from the source left.
        src_take = True  # Take the next source file from the generated files if true
        while True:
            try:
                if not src_done and src_take:
                    src_file = next(src_files)
            except StopIteration:
                src_file = None
                src_done = True

            try:
                if not dest_done and dest_take:
                    dest_file = next(dest_files)
            except StopIteration:
                dest_file = None
                dest_done = True

            if not (src_done or dest_done):
                src_take = True
                dest_take = True

                compare_keys = self.compare_comp_key(src_file, dest_file)
                if compare_keys == "equal":
                    should_sync = self._sync_strategy.determine_should_sync(src_file, dest_file)
                    if should_sync:
                        yield cast("FileStats", src_file)
                elif compare_keys == "less_than":
                    src_take = True
                    dest_take = False
                    should_sync = self._not_at_dest_sync_strategy.determine_should_sync(
                        src_file, None
                    )
                    if should_sync:
                        yield cast("FileStats", src_file)
                elif compare_keys == "greater_than":
                    src_take = False
                    dest_take = True
                    should_sync = self._not_at_src_sync_strategy.determine_should_sync(
                        None, dest_file
                    )
                    if should_sync:
                        yield cast("FileStats", dest_file)

            elif (not src_done) and dest_done:
                src_take = True
                should_sync = self._not_at_dest_sync_strategy.determine_should_sync(src_file, None)
                if should_sync:
                    yield cast("FileStats", src_file)
            elif src_done and (not dest_done):
                dest_take = True
                should_sync = self._not_at_src_sync_strategy.determine_should_sync(None, dest_file)
                if should_sync:
                    yield cast("FileStats", dest_file)
            else:
                break  # cov: ignore

    @staticmethod
    def compare_comp_key(src_file: FileStats | None, dest_file: FileStats | None) -> str:
        """Compare the source & destination compare_key."""
        src_comp_key = (src_file.compare_key if src_file else None) or ""
        dest_comp_key = (dest_file.compare_key if dest_file else None) or ""
        if src_comp_key == dest_comp_key:
            return "equal"
        if src_comp_key < dest_comp_key:
            return "less_than"
        return "greater_than"



"""CFNgin UI manipulation."""

from __future__ import annotations

import logging
import threading
from contextlib import AbstractContextManager
from getpass import getpass
from typing import TYPE_CHECKING, Any, TextIO

if TYPE_CHECKING:
    from types import TracebackType

    from typing_extensions import Self

LOGGER = logging.getLogger(__name__)


def get_raw_input(message: str) -> str:
    """Just a wrapper for :func:`input` for testing purposes."""
    return input(message)


class UI(AbstractContextManager["UI"]):
    """Used internally from terminal output in a multithreaded environment.

    Ensures that two threads don't write over each other while asking a user
    for input (e.g. in interactive mode).

    """

    def __init__(self) -> None:
        """Instantiate class."""
        self._lock = threading.RLock()

    def log(
        self,
        lvl: int,
        msg: Exception | str,
        *args: Any,
        logger: logging.Logger | logging.LoggerAdapter[Any] = LOGGER,
        **kwargs: Any,
    ) -> None:
        """Log the message if the current thread owns the underlying lock.

        Args:
            lvl: Log level.
            msg: String template or exception to use for the log record.
            logger: Specific logger to log to.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        kwargs["stacklevel"] = kwargs.get("stacklevel", 1) + 1
        with self:
            return logger.log(lvl, msg, *args, **kwargs)

    def info(
        self,
        msg: str,
        *args: Any,
        logger: logging.Logger | logging.LoggerAdapter[Any] = LOGGER,
        **kwargs: Any,
    ) -> None:
        """Log the line if the current thread owns the underlying lock.

        Args:
            msg: String template or exception to use
                for the log record.
            logger: Specific logger to log to.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        kwargs["logger"] = logger
        self.log(logging.INFO, msg, *args, **kwargs)

    def ask(self, message: str) -> str:
        """Collect input from a user in a multithreaded environment.

        This wraps the built-in input function to ensure that only 1
        thread is asking for input from the user at a give time. Any process
        that tries to log output to the terminal will be blocked while the
        user is being prompted.

        """
        with self:
            return get_raw_input(message)

    def getpass(self, prompt: str, stream: TextIO | None = None) -> str:
        """Wrap getpass to lock the UI."""
        with self:
            return getpass(prompt, stream)

    def __enter__(self) -> Self:
        """Enter the context manager."""
        self._lock.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit the context manager."""
        self._lock.__exit__(exc_type, exc_value, traceback)


# Global UI object for other modules to use.
ui = UI()



"""Transfer config.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/transferconfig.py

"""

from __future__ import annotations

from typing import Any, ClassVar, NoReturn

from s3transfer.manager import TransferConfig
from typing_extensions import TypedDict

from .utils import human_readable_to_bytes


# If the user does not specify any overrides,
# these are the default values we use for the s3 transfer
# commands.
class TransferConfigDict(TypedDict):
    max_bandwidth: int | str | None
    max_concurrent_requests: int
    max_queue_size: int
    multipart_chunksize: int | str
    multipart_threshold: int | str


DEFAULTS: TransferConfigDict = {
    "max_bandwidth": None,
    "max_concurrent_requests": 10,
    "max_queue_size": 1000,
    "multipart_chunksize": 8 * (1024**2),
    "multipart_threshold": 8 * (1024**2),
}


class InvalidConfigError(Exception):
    """Invalid configuration."""


class RuntimeConfig:
    """Runtime configuration."""

    POSITIVE_INTEGERS: ClassVar[list[str]] = [
        "max_bandwidth",
        "max_concurrent_requests",
        "max_queue_size",
        "multipart_chunksize",
        "multipart_threshold",
    ]
    HUMAN_READABLE_SIZES: ClassVar[list[str]] = [
        "multipart_chunksize",
        "multipart_threshold",
    ]
    HUMAN_READABLE_RATES: ClassVar[list[str]] = ["max_bandwidth"]

    @staticmethod
    def defaults() -> TransferConfigDict:
        """Return default values."""
        return DEFAULTS.copy()

    @classmethod
    def build_config(
        cls,
        *,
        max_bandwidth: int | str | None = None,
        max_concurrent_requests: int | str | None = None,
        max_queue_size: int | str | None = None,
        multipart_chunksize: int | str | None = None,
        multipart_threshold: int | str | None = None,
    ) -> TransferConfigDict:
        """Create and convert a runtime config dictionary.

        This method will merge and convert S3 runtime configuration
        data into a single dictionary that can then be passed to classes
        that use this runtime config.

        """
        runtime_config = DEFAULTS.copy()
        kwargs = {
            "max_bandwidth": max_bandwidth,
            "max_concurrent_requests": max_concurrent_requests,
            "max_queue_size": max_queue_size,
            "multipart_chunksize": multipart_chunksize,
            "multipart_threshold": multipart_threshold,
        }
        runtime_config.update({k: v for k, v in kwargs.items() if v is not None})  # type: ignore
        cls._convert_human_readable_sizes(runtime_config)
        cls._convert_human_readable_rates(runtime_config)
        cls._validate_config(runtime_config)
        return runtime_config

    @classmethod
    def _convert_human_readable_sizes(cls, runtime_config: TransferConfigDict) -> None:
        for attr in cls.HUMAN_READABLE_SIZES:
            value = runtime_config.get(attr)
            if isinstance(value, str):
                runtime_config[attr] = human_readable_to_bytes(value)

    @classmethod
    def _convert_human_readable_rates(cls, runtime_config: TransferConfigDict) -> None:
        for attr in cls.HUMAN_READABLE_RATES:
            value = runtime_config.get(attr)
            if isinstance(value, str):
                if not value.endswith("B/s"):
                    raise InvalidConfigError(
                        f"Invalid rate: {value}. The value must be expressed "
                        "as a rate in terms of bytes per seconds "
                        "(e.g. 10MB/s or 800KB/s)"
                    )
                runtime_config[attr] = human_readable_to_bytes(value[:-2])

    @classmethod
    def _validate_config(cls, runtime_config: TransferConfigDict) -> None:
        for attr in cls.POSITIVE_INTEGERS:
            value = runtime_config.get(attr)
            if value is not None:
                try:
                    runtime_config[attr] = int(value)
                    if not runtime_config[attr] > 0:
                        cls._error_positive_value(attr, value)
                except ValueError:
                    cls._error_positive_value(attr, value)

    @staticmethod
    def _error_positive_value(name: str, value: int) -> NoReturn:
        raise InvalidConfigError(f"Value for {name} must be a positive integer: {value}")


def create_transfer_config_from_runtime_config(
    runtime_config: TransferConfigDict,
) -> TransferConfig:
    """Create an equivalent s3transfer TransferConfig.

    Args:
        runtime_config: A valid RuntimeConfig-generated dict.

    Returns:
        A TransferConfig with the same configuration as the runtime config.

    """
    translation_map = {
        "max_bandwidth": "max_bandwidth",
        "max_concurrent_requests": "max_request_concurrency",
        "max_queue_size": "max_request_queue_size",
        "multipart_chunksize": "multipart_chunksize",
        "multipart_threshold": "multipart_threshold",
    }
    kwargs: dict[str, Any] = {}
    for key, value in runtime_config.items():
        if key not in translation_map:
            continue
        kwargs[translation_map[key]] = value
    return TransferConfig(**kwargs)



"""Runway AWS objects."""

from . import s3
from ._account import AccountDetails
from ._assume_role import AssumeRole
from ._response import BaseResponse, ResponseError, ResponseMetadata

__all__ = [
    "AccountDetails",
    "AssumeRole",
    "BaseResponse",
    "ResponseError",
    "ResponseMetadata",
    "s3",
]



"""CFNgin blueprint variable types."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeVar, overload

from troposphere import BaseAWSObject

if TYPE_CHECKING:
    from typing_extensions import Literal

TroposphereT = TypeVar("TroposphereT", bound=BaseAWSObject)


class TroposphereType(Generic[TroposphereT]):
    """Represents a Troposphere type.

    :class:`Troposphere` will convert the value provided to the variable to
    the specified Troposphere type.

    Both resource and parameter classes (which are just used to configure
    other resources) are acceptable as configuration values.

    Complete resource definitions must be dictionaries, with the keys
    identifying the resource titles, and the values being used as the
    constructor parameters.

    Parameter classes can be defined as dictionary or a list of
    dictionaries. In either case, the keys and values will be used directly
    as constructor parameters.

    """

    def __init__(
        self,
        defined_type: type[TroposphereT],
        *,
        many: bool = False,
        optional: bool = False,
        validate: bool = True,
    ) -> None:
        """Instantiate class.

        Args:
            defined_type: Troposphere type.
            many: Whether or not multiple resources can be constructed.
                If the defined type is a resource, multiple resources can be
                passed as a dictionary of dictionaries.
                If it is a parameter class, multiple resources are passed as
                a list.
            optional: Whether an undefined/null configured value is acceptable.
                In that case a value of ``None`` will be passed to the template,
                even if ``many`` is enabled.
            validate: Whether to validate the generated object on creation.
                Should be left enabled unless the object will be augmented with
                mandatory parameters in the template code, such that it must be
                validated at a later point.

        """
        self._validate_type(defined_type)

        self._type = defined_type
        self._many = many
        self._optional = optional
        self._validate = validate

    @staticmethod
    def _validate_type(defined_type: type[TroposphereT]) -> None:
        if not hasattr(defined_type, "from_dict"):
            raise ValueError("Type must have `from_dict` attribute")

    @property
    def resource_name(self) -> str:
        """Name of the type or resource."""
        return str(getattr(self._type, "resource_name", None) or self._type.__name__)

    @overload
    def create(self, value: dict[str, Any]) -> TroposphereT: ...

    @overload
    def create(self, value: list[dict[str, Any]]) -> list[TroposphereT]: ...

    @overload
    def create(self, value: None) -> None: ...

    def create(
        self, value: dict[str, Any] | list[dict[str, Any]] | None
    ) -> TroposphereT | list[TroposphereT] | None:
        """Create the troposphere type from the value.

        Args:
            value: A dictionary or list of dictionaries (see class documentation
                for details) to use as parameters to create the Troposphere type instance.
                Each dictionary will be passed to the ``from_dict`` method of
                the type.

        Returns:
            Returns the value converted to the troposphere type.

        """
        # Explicitly check with len such that non-sequence types throw.
        if self._optional and (value is None or len(value) == 0):
            return None

        if hasattr(self._type, "resource_type"):
            # Our type is a resource, so ensure we have a dict of title to
            # parameters
            if not isinstance(value, dict):
                raise ValueError("Resources must be specified as a dict of title to parameters")
            if not self._many and len(value) > 1:
                raise ValueError(
                    "Only one resource can be provided for this TroposphereType variable"
                )

            result = [self._type.from_dict(title, v) for title, v in value.items()]
        elif self._many and isinstance(value, list):
            result = [self._type.from_dict(None, v) for v in value]
        elif not isinstance(value, dict):
            raise ValueError(
                "TroposphereType for a single non-resource"
                "type must be specified as a dict of "
                "parameters"
            )
        else:
            result = [self._type.from_dict(None, value)]

        if self._validate:
            for v in result:
                v._validate_props()  # noqa: SLF001

        return result[0] if not self._many else result


class CFNType:
    """Represents a CloudFormation Parameter Type.

    :class:`CFNType` can be used as the ``type`` for a Blueprint variable.
    Unlike other variables, a variable with ``type: CFNType``, will
    be submitted to CloudFormation as a Parameter.

    Attributes:
        parameter_type: Name of the CloudFormation Parameter type to specify when
            submitting as a CloudFormation Parameter.

    See Also:
        https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html

    """

    parameter_type: ClassVar


# General CFN types
class CFNString(CFNType):
    """A literal string."""

    parameter_type: ClassVar[Literal["String"]] = "String"


class CFNNumber(CFNType):
    """An integer or float.

    AWS CloudFormation validates the parameter value as a number; however,
    when you use the parameter elsewhere in your template (for example, by using
    the Ref intrinsic function), the parameter value becomes a string.

    """

    parameter_type: ClassVar[Literal["Number"]] = "Number"


class CFNNumberList(CFNType):
    """An array of integers or floats that are separated by commas.

    AWS CloudFormation validates the parameter value as numbers; however,
    when you use the parameter elsewhere in your template (for example, by using
    the Ref intrinsic function), the parameter value becomes a list of strings.

    """

    parameter_type: ClassVar[Literal["List<Number>"]] = "List<Number>"


class CFNCommaDelimitedList(CFNType):
    """An array of literal strings that are separated by commas.

    The total number of strings should be one more than the total number of commas.
    Also, each member string is space trimmed.

    """

    parameter_type: ClassVar[Literal["CommaDelimitedList"]] = "CommaDelimitedList"


# AWS-Specific Parameter Types
# https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html#aws-specific-parameter-types
class EC2AvailabilityZoneName(CFNType):
    """An Availability Zone, such as us-west-2a."""

    parameter_type: ClassVar[Literal["AWS::EC2::AvailabilityZone::Name"]] = (
        "AWS::EC2::AvailabilityZone::Name"
    )


class EC2ImageId(CFNType):
    """An Amazon EC2 image ID, such as ami-0ff8a91507f77f867.

    Note that the AWS CloudFormation console doesn't show a drop-down list of
    values for this parameter type.

    """

    parameter_type: ClassVar[Literal["AWS::EC2::Image::Id"]] = "AWS::EC2::Image::Id"


class EC2InstanceId(CFNType):
    """An Amazon EC2 instance ID, such as i-1e731a32."""

    parameter_type: ClassVar[Literal["AWS::EC2::Instance::Id"]] = "AWS::EC2::Instance::Id"


class EC2KeyPairKeyName(CFNType):
    """An Amazon EC2 key pair name."""

    parameter_type: ClassVar[Literal["AWS::EC2::KeyPair::KeyName"]] = "AWS::EC2::KeyPair::KeyName"


class EC2SecurityGroupGroupName(CFNType):
    """An EC2-Classic or default VPC security group name, such as my-sg-abc."""

    parameter_type: ClassVar[Literal["AWS::EC2::SecurityGroup::GroupName"]] = (
        "AWS::EC2::SecurityGroup::GroupName"
    )


class EC2SecurityGroupId(CFNType):
    """A security group ID, such as sg-a123fd85."""

    parameter_type: ClassVar[Literal["AWS::EC2::SecurityGroup::Id"]] = "AWS::EC2::SecurityGroup::Id"


class EC2SubnetId(CFNType):
    """A subnet ID, such as subnet-123a351e."""

    parameter_type: ClassVar[Literal["AWS::EC2::Subnet::Id"]] = "AWS::EC2::Subnet::Id"


class EC2VolumeId(CFNType):
    """An Amazon EBS volume ID, such as vol-3cdd3f56."""

    parameter_type: ClassVar[Literal["AWS::EC2::Volume::Id"]] = "AWS::EC2::Volume::Id"


class EC2VPCId(CFNType):
    """A VPC ID, such as vpc-a123baa3."""

    parameter_type: ClassVar[Literal["AWS::EC2::VPC::Id"]] = "AWS::EC2::VPC::Id"


class Route53HostedZoneId(CFNType):
    """An Amazon Route 53 hosted zone ID, such as Z23YXV4OVPL04A."""

    parameter_type: ClassVar[Literal["AWS::Route53::HostedZone::Id"]] = (
        "AWS::Route53::HostedZone::Id"
    )


class EC2AvailabilityZoneNameList(CFNType):
    """An array of Availability Zones for a region, such as us-west-2a, us-west-2b."""

    parameter_type: ClassVar[Literal["List<AWS::EC2::AvailabilityZone::Name>"]] = (
        "List<AWS::EC2::AvailabilityZone::Name>"
    )


class EC2ImageIdList(CFNType):
    """An array of Amazon EC2 image IDs, such as ami-0ff8a91507f77f867, ami-0a584ac55a7631c0c.

    Note that the AWS CloudFormation console doesn't show a drop-down list of
    values for this parameter type.

    """

    parameter_type: ClassVar[Literal["List<AWS::EC2::Image::Id>"]] = "List<AWS::EC2::Image::Id>"


class EC2InstanceIdList(CFNType):
    """An array of Amazon EC2 instance IDs, such as i-1e731a32, i-1e731a34."""

    parameter_type: ClassVar[Literal["List<AWS::EC2::Instance::Id>"]] = (
        "List<AWS::EC2::Instance::Id>"
    )


class EC2SecurityGroupGroupNameList(CFNType):
    """An array of EC2-Classic or default VPC security group names."""

    parameter_type: ClassVar[Literal["List<AWS::EC2::SecurityGroup::GroupName>"]] = (
        "List<AWS::EC2::SecurityGroup::GroupName>"
    )


class EC2SecurityGroupIdList(CFNType):
    """An array of security group IDs, such as sg-a123fd85, sg-b456fd85."""

    parameter_type: ClassVar[Literal["List<AWS::EC2::SecurityGroup::Id>"]] = (
        "List<AWS::EC2::SecurityGroup::Id>"
    )


class EC2SubnetIdList(CFNType):
    """An array of subnet IDs, such as subnet-123a351e, subnet-456b351e."""

    parameter_type: ClassVar[Literal["List<AWS::EC2::Subnet::Id>"]] = "List<AWS::EC2::Subnet::Id>"


class EC2VolumeIdList(CFNType):
    """An array of Amazon EBS volume IDs, such as vol-3cdd3f56, vol-4cdd3f56."""

    parameter_type: ClassVar[Literal["List<AWS::EC2::Volume::Id>"]] = "List<AWS::EC2::Volume::Id>"


class EC2VPCIdList(CFNType):
    """An array of VPC IDs, such as vpc-a123baa3, vpc-b456baa3."""

    parameter_type: ClassVar[Literal["List<AWS::EC2::VPC::Id>"]] = "List<AWS::EC2::VPC::Id>"


class Route53HostedZoneIdList(CFNType):
    """An array of Amazon Route 53 hosted zone IDs, such as Z23YXV4OVPL04A, Z23YXV4OVPL04B."""

    parameter_type: ClassVar[Literal["List<AWS::Route53::HostedZone::Id>"]] = (
        "List<AWS::Route53::HostedZone::Id>"
    )


# SSM Parameter Types
# https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html#aws-ssm-parameter-types
class SSMParameterName(CFNType):
    """The name of a Systems Manager parameter key.

    Use this parameter when you want to pass the parameter key.
    For example, you can use this type to validate that the parameter exists.

    """

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Name"]] = "AWS::SSM::Parameter::Name"


class SSMParameterValueString(CFNType):
    """A Systems Manager parameter whose value is a string.

    This corresponds to the String parameter type in Parameter Store.

    """

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<String>"]] = (
        "AWS::SSM::Parameter::Value<String>"
    )


class SSMParameterValueStringList(CFNType):
    """A Systems Manager parameter whose value is a list of strings.

    This corresponds to the StringList parameter type in Parameter Store.

    """

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<List<String>>"]] = (
        "AWS::SSM::Parameter::Value<List<String>>"
    )


class SSMParameterValueCommaDelimitedList(CFNType):
    """A Systems Manager parameter whose value is a list of strings.

    This corresponds to the StringList parameter type in Parameter Store.

    """

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<CommaDelimitedList>"]] = (
        "AWS::SSM::Parameter::Value<CommaDelimitedList>"
    )


class SSMParameterValueEC2AvailabilityZoneName(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[
        Literal["AWS::SSM::Parameter::Value<AWS::EC2::AvailabilityZone::Name>"]
    ] = "AWS::SSM::Parameter::Value<AWS::EC2::AvailabilityZone::Name>"


class SSMParameterValueEC2ImageId(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>"]] = (
        "AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>"
    )


class SSMParameterValueEC2InstanceId(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<AWS::EC2::Instance::Id>"]] = (
        "AWS::SSM::Parameter::Value<AWS::EC2::Instance::Id>"
    )


class SSMParameterValueEC2KeyPairKeyName(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<AWS::EC2::KeyPair::KeyName>"]] = (
        "AWS::SSM::Parameter::Value<AWS::EC2::KeyPair::KeyName>"
    )


class SSMParameterValueEC2SecurityGroupGroupName(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[
        Literal["AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::GroupName>"]
    ] = "AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::GroupName>"


class SSMParameterValueEC2SecurityGroupId(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::Id>"]] = (
        "AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::Id>"
    )


class SSMParameterValueEC2SubnetId(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<AWS::EC2::Subnet::Id>"]] = (
        "AWS::SSM::Parameter::Value<AWS::EC2::Subnet::Id>"
    )


class SSMParameterValueEC2VolumeId(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<AWS::EC2::Volume::Id>"]] = (
        "AWS::SSM::Parameter::Value<AWS::EC2::Volume::Id>"
    )


class SSMParameterValueEC2VPCId(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<AWS::EC2::VPC::Id>"]] = (
        "AWS::SSM::Parameter::Value<AWS::EC2::VPC::Id>"
    )


class SSMParameterValueRoute53HostedZoneId(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[
        Literal["AWS::SSM::Parameter::Value<AWS::Route53::HostedZone::Id>"]
    ] = "AWS::SSM::Parameter::Value<AWS::Route53::HostedZone::Id>"


class SSMParameterValueEC2AvailabilityZoneNameList(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[
        Literal["AWS::SSM::Parameter::Value<List<AWS::EC2::AvailabilityZone::Name>>"]
    ] = "AWS::SSM::Parameter::Value<List<AWS::EC2::AvailabilityZone::Name>>"


class SSMParameterValueEC2ImageIdList(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<List<AWS::EC2::Image::Id>>"]] = (
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Image::Id>>"
    )


class SSMParameterValueEC2InstanceIdList(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[
        Literal["AWS::SSM::Parameter::Value<List<AWS::EC2::Instance::Id>>"]
    ] = "AWS::SSM::Parameter::Value<List<AWS::EC2::Instance::Id>>"


class SSMParameterValueEC2SecurityGroupGroupNameList(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[
        Literal["AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::GroupName>>"]
    ] = "AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::GroupName>>"


class SSMParameterValueEC2SecurityGroupIdList(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[
        Literal["AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::Id>>"]
    ] = "AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::Id>>"


class SSMParameterValueEC2SubnetIdList(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<List<AWS::EC2::Subnet::Id>>"]] = (
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Subnet::Id>>"
    )


class SSMParameterValueEC2VolumeIdList(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<List<AWS::EC2::Volume::Id>>"]] = (
        "AWS::SSM::Parameter::Value<List<AWS::EC2::Volume::Id>>"
    )


class SSMParameterValueEC2VPCIdList(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[Literal["AWS::SSM::Parameter::Value<List<AWS::EC2::VPC::Id>>"]] = (
        "AWS::SSM::Parameter::Value<List<AWS::EC2::VPC::Id>>"
    )


class SSMParameterValueRoute53HostedZoneIdList(CFNType):
    """A Systems Manager parameter whose value is an AWS-specific parameter type."""

    parameter_type: ClassVar[
        Literal["AWS::SSM::Parameter::Value<List<AWS::Route53::HostedZone::Id>>"]
    ] = "AWS::SSM::Parameter::Value<List<AWS::Route53::HostedZone::Id>>"



"""AWS SSM Parameter Store hooks."""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, Annotated, Any, ClassVar, cast

from pydantic import ConfigDict, Field, field_validator
from typing_extensions import Literal, TypedDict

from ....compat import cached_property
from ....utils import BaseModel, JsonEncoder
from ..protocols import CfnginHookProtocol
from ..utils import TagDataModel

if TYPE_CHECKING:
    from mypy_boto3_ssm.client import SSMClient
    from mypy_boto3_ssm.literals import ParameterTierType
    from mypy_boto3_ssm.type_defs import ParameterTypeDef, TagTypeDef

    from ...._logging import RunwayLogger
    from ....context import CfnginContext
else:
    ParameterTierType = Literal["Advanced", "Intelligent-Tiering", "Standard"]

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


# PutParameterResultTypeDef but without metadata
class _PutParameterResultTypeDef(TypedDict):
    Tier: ParameterTierType
    Version: int


class ArgsDataModel(BaseModel):
    """Parameter hook args."""

    model_config = ConfigDict(extra="ignore", populate_by_name=True)

    allowed_pattern: Annotated[str | None, Field(alias="AllowedPattern")] = None
    """A regular expression used to validate the parameter value."""

    data_type: Annotated[
        Literal["aws:ec2:image", "text"] | None,
        Field(alias="DataType"),
    ] = None
    """The data type for a String parameter.

    Supported data types include plain text and Amazon Machine Image IDs.

    """

    description: Annotated[str | None, Field(alias="Description")] = None
    """Information about the parameter."""

    force: bool = False
    """Skip checking the current value of the parameter, just put it.

    Can be used alongside ``overwrite`` to always update a parameter.

    """

    key_id: Annotated[str | None, Field(alias="KeyId")] = None
    """The KMS Key ID that you want to use to encrypt a parameter.

    Either the default AWS Key Management Service (AWS KMS) key automatically
    assigned to your AWS account or a custom key.
    Required for parameters that use the ``SecureString`` data type.

    """

    name: Annotated[str, Field(alias="Name")]
    """The fully qualified name of the parameter that you want to add to the system."""

    overwrite: Annotated[bool, Field(alias="Overwrite")] = True
    """Allow overwriting an existing parameter."""

    policies: Annotated[str | None, Field(alias="Policies")] = None
    """One or more policies to apply to a parameter. This field takes a JSON array."""

    tags: Annotated[list[TagDataModel] | None, Field(alias="Tags")] = None
    """Optional metadata that you assign to a resource."""

    tier: Annotated[ParameterTierType, Field(alias="Tier")] = "Standard"
    """The parameter tier to assign to a parameter."""

    type: Annotated[Literal["String", "StringList", "SecureString"], Field(alias="Type")]
    """The type of parameter."""

    value: Annotated[str | None, Field(alias="Value")] = None
    """The parameter value that you want to add to the system.

    Standard parameters have a value limit of 4 KB.
    Advanced parameters have a value limit of 8 KB.

    """

    @field_validator("policies", mode="before")
    @classmethod
    def _convert_policies(cls, v: list[dict[str, Any]] | str | Any) -> str:
        """Convert policies to acceptable value."""
        if isinstance(v, str):
            return v
        if isinstance(v, list):
            return json.dumps(v, cls=JsonEncoder)
        raise ValueError(f"unexpected type {type(v)}; permitted: list[dict[str, Any]] | str | None")

    @field_validator("tags", mode="before")
    @classmethod
    def _convert_tags(cls, v: dict[str, str] | list[dict[str, str]] | Any) -> list[dict[str, str]]:
        """Convert tags to acceptable value."""
        if isinstance(v, list):
            return v
        if isinstance(v, dict):
            return [{"Key": k, "Value": v} for k, v in v.items()]
        raise ValueError(
            f"unexpected type {type(v)}; permitted: dict[str, str] | list[dict[str, str]] | none"
        )


class _Parameter(CfnginHookProtocol):
    """AWS SSM Parameter Store Parameter."""

    ARGS_PARSER: ClassVar = ArgsDataModel
    """Class used to parse arguments passed to the hook."""

    args: ArgsDataModel

    def __init__(
        self,
        context: CfnginContext,
        *,
        name: str,
        type: Literal["String", "StringList", "SecureString"],  # noqa: A002
        **kwargs: Any,
    ) -> None:
        """Instantiate class.

        Args:
            context: CFNgin context object.
            name: The fully qualified name of the parameter that you want to add to
                the system.
            type: The type of parameter.
            **kwargs: Arbitrary keyword arguments.

        """
        self.args = ArgsDataModel.model_validate({"name": name, "type": type, **kwargs})
        self.ctx = context

    @cached_property
    def client(self) -> SSMClient:
        """AWS SSM client."""
        return self.ctx.get_session().client("ssm")

    def delete(self) -> bool:
        """Delete parameter."""
        try:
            self.client.delete_parameter(Name=self.args.name)
            LOGGER.info("deleted SSM Parameter %s", self.args.name)
        except self.client.exceptions.ParameterNotFound:
            LOGGER.info("delete parameter skipped; %s not found", self.args.name)
        return True

    def get(self) -> ParameterTypeDef:
        """Get parameter."""
        if self.args.force:  # bypass getting current value
            return {}
        try:
            return self.client.get_parameter(Name=self.args.name, WithDecryption=True).get(
                "Parameter", {}
            )
        except self.client.exceptions.ParameterNotFound:
            LOGGER.verbose("parameter %s does not exist", self.args.name)
            return {}

    def get_current_tags(self) -> list[TagTypeDef]:
        """Get Tags currently applied to Parameter."""
        try:
            return self.client.list_tags_for_resource(
                ResourceId=self.args.name, ResourceType="Parameter"
            ).get("TagList", [])
        except (
            self.client.exceptions.InvalidResourceId,
            self.client.exceptions.ParameterNotFound,
        ):
            return []

    def post_deploy(self) -> _PutParameterResultTypeDef:
        """Run during the *post_deploy* stage."""
        result = self.put()
        self.update_tags()
        return result

    def post_destroy(self) -> bool:
        """Run during the *post_destroy* stage."""
        return self.delete()

    def pre_deploy(self) -> _PutParameterResultTypeDef:
        """Run during the *pre_deploy* stage."""
        result = self.put()
        self.update_tags()
        return result

    def pre_destroy(self) -> bool:
        """Run during the *pre_destroy* stage."""
        return self.delete()

    def put(self) -> _PutParameterResultTypeDef:
        """Put parameter."""
        if not self.args.value:
            LOGGER.info(
                "skipped putting SSM Parameter; value provided for %s is falsy",
                self.args.name,
            )
            return {"Tier": self.args.tier, "Version": 0}
        current_param = self.get()
        if current_param.get("Value") != self.args.value:
            try:
                result = self.client.put_parameter(
                    **self.args.model_dump(
                        by_alias=True, exclude_none=True, exclude={"force", "tags"}
                    )
                )
            except self.client.exceptions.ParameterAlreadyExists:
                LOGGER.warning(
                    "parameter %s already exists; to overwrite it's value, "
                    'set the overwrite field to "true"',
                    self.args.name,
                )
                return {
                    "Tier": current_param.get("Tier", self.args.tier),
                    "Version": current_param.get("Version", 0),
                }
        else:
            result: _PutParameterResultTypeDef = {
                "Tier": current_param.get("Tier", self.args.tier),
                "Version": current_param.get("Version", 0),
            }
        LOGGER.info("put SSM Parameter %s", self.args.name)
        return result

    def update_tags(self) -> None:
        """Update tags."""
        current_tags = self.get_current_tags()
        if self.args.tags and current_tags:
            diff_tag_keys = list({i["Key"] for i in current_tags} ^ {i.key for i in self.args.tags})
        elif self.args.tags:
            diff_tag_keys = []
        else:
            diff_tag_keys = [i["Key"] for i in current_tags]

        try:
            if diff_tag_keys:
                diff_tag_keys.sort()
                self.client.remove_tags_from_resource(
                    ResourceId=self.args.name,
                    ResourceType="Parameter",
                    TagKeys=diff_tag_keys,
                )
                LOGGER.debug("removed tags for parameter %s: %s", self.args.name, diff_tag_keys)

            if self.args.tags:
                tags_to_add = [
                    cast("TagTypeDef", tag.model_dump(by_alias=True)) for tag in self.args.tags
                ]
                self.client.add_tags_to_resource(
                    ResourceId=self.args.name,
                    ResourceType="Parameter",
                    Tags=tags_to_add,
                )
                LOGGER.debug(
                    "added tags to parameter %s: %s",
                    self.args.name,
                    [tag["Key"] for tag in tags_to_add],
                )
        except self.client.exceptions.InvalidResourceId:
            LOGGER.info("skipped updating tags; parameter %s does not exist", self.args.name)
        else:
            LOGGER.info("updated tags for parameter %s", self.args.name)


class SecureString(_Parameter):
    """AWS SSM Parameter Store SecureString Parameter."""

    def __init__(
        self,
        context: CfnginContext,
        *,
        name: str,
        **kwargs: Any,
    ) -> None:
        """Instantiate class.

        Args:
            context: CFNgin context object.
            name: The fully qualified name of the parameter that you want to add to
                the system.
            **kwargs: Arbitrary keyword arguments.

        """
        for k in ["Type", "type"]:  # ensure neither of these are set
            kwargs.pop(k, None)
        super().__init__(context, name=name, type="SecureString", **kwargs)



"""Hook data lookup."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar

from troposphere import BaseAWSObject

from ....lookups.handlers.base import LookupHandler
from ....utils import MutableMap

if TYPE_CHECKING:

    from ....context import CfnginContext

LOGGER = logging.getLogger(__name__)


class HookDataLookup(LookupHandler["CfnginContext"]):
    """Hook data lookup."""

    TYPE_NAME: ClassVar[str] = "hook_data"
    """Name that the Lookup is registered as."""

    @classmethod
    def handle(cls, value: str, context: CfnginContext, **_: Any) -> Any:
        """Return the data from ``hook_data``.

        Args:
            value: Parameter(s) given to this lookup.
            context: Context instance.

        """
        query, args = cls.parse(value)

        hook_data = MutableMap(**context.hook_data)

        result = hook_data.find(query, args.get("default"))

        if isinstance(result, BaseAWSObject) and args.get("get") and not args.get("load"):
            args["load"] = "troposphere"

        if not result:
            raise ValueError(f'Could not find a value for "{value}"')

        if result == args.get("default"):
            # assume default value has already been processed so no need to
            # use these
            args.pop("load", None)
            args.pop("get", None)

        return cls.format_results(result, **args)



"""Hook for creating an AWS Lambda Function using Python runtime."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar

from ....compat import cached_property
from .base_classes import AwsLambdaHook
from .models.args import PythonHookArgs
from .python_requirements import PythonDeploymentPackage, PythonProject

if TYPE_CHECKING:
    from ....context import CfnginContext
    from .base_classes import DeploymentPackage

LOGGER = logging.getLogger(__name__.replace("._", "."))


class PythonFunction(AwsLambdaHook[PythonProject]):
    """Hook for creating an AWS Lambda Function using Python runtime."""

    BUILD_LAYER: ClassVar[bool] = False
    """Flag to denote that this hook creates a Lambda Function deployment package."""

    args: PythonHookArgs
    """Parsed hook arguments."""

    def __init__(self, context: CfnginContext, **kwargs: Any) -> None:
        """Instantiate class."""
        super().__init__(context)
        self.args = PythonHookArgs.model_validate(kwargs)

    @cached_property
    def deployment_package(self) -> DeploymentPackage[PythonProject]:
        """AWS Lambda deployment package."""
        return PythonDeploymentPackage.init(
            self.project, "layer" if self.BUILD_LAYER else "function"
        )

    @cached_property
    def project(self) -> PythonProject:
        """Project being deployed as an AWS Lambda Function."""
        return PythonProject(self.args, self.ctx)

    def cleanup(self) -> None:
        """Cleanup after execution."""
        self.project.cleanup()

    def cleanup_on_error(self) -> None:
        """Cleanup after an error has occurred."""
        self.deployment_package.delete()
        self.project.cleanup_on_error()

    def pre_deploy(self) -> Any:
        """Run during the **pre_deploy** stage."""
        try:
            self.deployment_package.upload()
            return self.build_response("deploy").model_dump(by_alias=True)
        except BaseException:
            self.cleanup_on_error()
            raise
        finally:
            self.cleanup()


class PythonLayer(PythonFunction):
    """Hook for creating an AWS Lambda Layer using Python runtime."""

    BUILD_LAYER: ClassVar[bool] = True
    """Flag to denote that this hook creates a Lambda Layer deployment package."""



"""Python project."""

from __future__ import annotations

import logging
import shutil
from typing import TYPE_CHECKING, ClassVar

from .....compat import cached_property
from .....dependency_managers import (
    Pip,
    Pipenv,
    PipenvNotFoundError,
    Poetry,
    PoetryNotFoundError,
)
from ..base_classes import Project
from ..models.args import PythonHookArgs
from . import PythonDockerDependencyInstaller

if TYPE_CHECKING:
    from pathlib import Path

    from typing_extensions import Literal

LOGGER = logging.getLogger(__name__.replace("._", "."))


class PythonProject(Project[PythonHookArgs]):
    """Python project."""

    DEFAULT_CACHE_DIR_NAME: ClassVar[str] = "pip_cache"
    """Name of the default cache directory."""

    @cached_property
    def docker(self) -> PythonDockerDependencyInstaller | None:
        """Docker interface that can be used to build the project."""
        return PythonDockerDependencyInstaller.from_project(self)

    @cached_property
    def metadata_files(self) -> tuple[Path, ...]:
        """Project metadata files.

        Files are only included in return value if they exist.

        """
        if self.project_type == "poetry":
            config_files = [self.project_root / config_file for config_file in Poetry.CONFIG_FILES]
        elif self.project_type == "pipenv":
            config_files = [self.project_root / config_file for config_file in Pipenv.CONFIG_FILES]
        else:
            config_files = [self.project_root / config_file for config_file in Pip.CONFIG_FILES]
        return tuple(path for path in config_files if path.exists())

    @cached_property
    def runtime(self) -> str:
        """Runtime of the build system.

        Value should be a valid Lambda Function runtime
        (https://docs.aws.amazon.com/lambda/latest/dg/lambda-runtimes.html).

        """
        if self._runtime_from_docker:
            return self._validate_runtime(self._runtime_from_docker)
        return self._validate_runtime(
            f"python{self.pip.python_version.major}.{self.pip.python_version.minor}"
        )

    @cached_property
    def pip(self) -> Pip:
        """Pip dependency manager."""
        return Pip(self.ctx, self.project_root)

    @cached_property
    def pipenv(self) -> Pipenv | None:
        """Pipenv dependency manager.

        Return:
            If the project uses pipenv and pipenv is not explicitly disabled,
            an object for interfacing with pipenv will be returned.

        Raises:
            PipenvNotFoundError: pipenv is not installed or not found in PATH.

        """
        if self.project_type != "pipenv":
            return None
        if Pipenv.found_in_path():
            return Pipenv(self.ctx, self.project_root)
        raise PipenvNotFoundError

    @cached_property
    def poetry(self) -> Poetry | None:
        """Poetry dependency manager.

        Return:
            If the project uses poetry and poetry is not explicitly disabled,
            an object for interfacing with poetry will be returned.

        Raises:
            PoetryNotFound: poetry is not installed or not found in PATH.

        """
        if self.project_type != "poetry":
            return None
        if Poetry.found_in_path():
            return Poetry(self.ctx, self.project_root)
        raise PoetryNotFoundError

    @cached_property
    def project_type(self) -> Literal["pip", "pipenv", "poetry"]:
        """Type of python project."""
        if Poetry.dir_is_project(self.project_root):
            if self.args.use_poetry:
                return "poetry"
            LOGGER.warning("poetry project detected but use of poetry is explicitly disabled")
        if Pipenv.dir_is_project(self.project_root):
            if self.args.use_pipenv:
                return "pipenv"
            LOGGER.warning("pipenv project detected but use of pipenv is explicitly disabled")
        return "pip"

    @cached_property
    def requirements_txt(self) -> Path | None:
        """Dependency file for the project."""
        if self.poetry:  # prioritize poetry
            return self.poetry.export(output=self.tmp_requirements_txt)
        if self.pipenv:
            return self.pipenv.export(output=self.tmp_requirements_txt)
        requirements_txt = self.project_root / "requirements.txt"
        if Pip.dir_is_project(self.project_root, file_name=requirements_txt.name):
            return requirements_txt
        return None

    @cached_property
    def supported_metadata_files(self) -> set[str]:
        """Names of all supported metadata files.

        Returns:
            Set of file names - not paths.

        """
        file_names = {*Pip.CONFIG_FILES}
        if self.args.use_poetry:
            file_names.update(Poetry.CONFIG_FILES)
        if self.args.use_pipenv:
            file_names.update(Pipenv.CONFIG_FILES)
        return file_names

    @cached_property
    def tmp_requirements_txt(self) -> Path:
        """Temporary requirements.txt file.

        This path is only used when exporting from another format.

        """
        return self.ctx.work_dir / f"{self.source_code.md5_hash}.requirements.txt"

    def cleanup(self) -> None:
        """Cleanup temporary files after the build process has run."""
        if (self.poetry or self.pipenv) and self.tmp_requirements_txt.exists():
            self.tmp_requirements_txt.unlink()
        shutil.rmtree(self.dependency_directory, ignore_errors=True)
        if not any(self.build_directory.iterdir()):
            # remove build_directory if it's empty
            shutil.rmtree(self.build_directory, ignore_errors=True)

    def install_dependencies(self) -> None:
        """Install project dependencies."""
        if self.requirements_txt:
            LOGGER.debug("installing dependencies to %s...", self.dependency_directory)
            if self.docker:
                self.docker.install()
            else:
                self.pip.install(
                    cache_dir=self.args.cache_dir,
                    extend_args=self.args.extend_pip_args,
                    no_cache_dir=not self.args.use_cache,
                    no_deps=bool(self.poetry or self.pipenv),
                    requirements=self.requirements_txt,
                    target=self.dependency_directory,
                )
            LOGGER.debug("dependencies successfully installed to %s", self.dependency_directory)
        else:
            LOGGER.info("skipped installing dependencies; none found")



"""Runway test definition models."""

from __future__ import annotations

from typing import Annotated, Any

from pydantic import ConfigDict, Field, field_validator
from typing_extensions import Literal

from .. import utils
from ..base import ConfigProperty

ValidRunwayTestTypeValues = Literal["cfn-lint", "script", "yamllint"]


class RunwayTestDefinitionModel(ConfigProperty):
    """Model for a Runway test definition."""

    model_config = ConfigDict(
        extra="ignore",
        json_schema_extra={
            "description": "Tests that can be run via the 'test' command.",
        },
        title="Runway Test Definition",
        use_enum_values=True,
        validate_assignment=True,
        validate_default=True,
    )

    args: Annotated[
        dict[str, Any] | str,
        Field(
            title="Arguments",
            description="Arguments to be passed to the test. Support varies by test type.",
        ),
    ] = {}
    name: Annotated[str, Field(description="Name of the test.")] = "test-name"
    required: Annotated[
        bool | str, Field(description="Whether the test must pass for subsequent tests to be run.")
    ] = False
    type: ValidRunwayTestTypeValues

    # TODO (kyle): add regex to schema
    _validate_string_is_lookup = field_validator("args", "required", mode="before")(
        utils.validate_string_is_lookup
    )



#!/usr/bin/env python
"""Module with static website bucket and CloudFront distribution."""
from __future__ import annotations

import hashlib
import logging
import os
from typing import TYPE_CHECKING, Any, ClassVar

import awacs.awslambda
import awacs.iam
import awacs.logs
import awacs.s3
import awacs.states
import awacs.sts
from awacs.aws import Action, Allow, Policy, PolicyDocument, Principal, Statement
from awacs.helpers.trust import make_simple_assume_policy
from troposphere import (
    AccountId,
    Join,
    NoValue,
    Output,
    Partition,
    Region,
    StackName,
    awslambda,
    cloudfront,
    iam,
    s3,
)
from typing_extensions import TypedDict

from ...cfngin.blueprints.base import Blueprint
from ...context import CfnginContext

if TYPE_CHECKING:
    from troposphere import Ref

    from ...cfngin.blueprints.type_defs import BlueprintVariableTypeDef

LOGGER = logging.getLogger("runway")

IAM_ARN_PREFIX = "arn:aws:iam::aws:policy/service-role/"


class _IndexRewriteFunctionInfoTypeDef(TypedDict):
    function: awslambda.Function
    role: iam.Role
    version: awslambda.Version


class StaticSite(Blueprint):
    """CFNgin blueprint for creating S3 bucket and CloudFront distribution."""

    VARIABLES: ClassVar[dict[str, BlueprintVariableTypeDef]] = {
        "AcmCertificateArn": {
            "type": str,
            "default": "",
            "description": "(Optional) Cert ARN for site",
        },
        "Aliases": {
            "type": list,
            "default": [],
            "description": "(Optional) Domain aliases the distribution",
        },
        "Compress": {
            "type": bool,
            "default": True,
            "description": "Whether the CloudFront default cache behavior will "
            "automatically compress certain files.",
        },
        "DisableCloudFront": {
            "type": bool,
            "default": False,
            "description": "Whether to disable CF",
        },
        "LogBucketName": {
            "type": str,
            "default": "",
            "description": "S3 bucket for CF logs",
        },
        "PriceClass": {
            "type": str,
            "default": "PriceClass_100",  # US/Europe
            "description": "CF price class for the distribution.",
        },
        "RewriteDirectoryIndex": {
            "type": str,
            "default": "",
            "description": "(Optional) File name to append to directory requests.",
        },
        "RoleBoundaryArn": {
            "type": str,
            "default": "",
            "description": "(Optional) IAM Role permissions "
            "boundary applied to any created "
            "roles.",
        },
        "WAFWebACL": {
            "type": str,
            "default": "",
            "description": "(Optional) WAF id to associate with the distribution.",
        },
        "custom_error_responses": {
            "type": list,
            "default": [],
            "description": "(Optional) Custom error responses.",
        },
        "lambda_function_associations": {
            "type": list,
            "default": [],
            "description": "(Optional) Lambda function associations.",
        },
    }

    @property
    def aliases_specified(self) -> bool:
        """Aliases are specified conditional."""
        return self.variables["Aliases"] != [""]

    @property
    def cf_enabled(self) -> bool:
        """CloudFront enabled conditional."""
        return not self.variables.get("DisableCloudFront", False)

    @property
    def acm_certificate_specified(self) -> bool:
        """ACM Certification specified conditional."""
        return self.variables["AcmCertificateArn"] != ""

    @property
    def cf_logging_enabled(self) -> bool:
        """CloudFront Logging specified conditional."""
        return self.variables["LogBucketName"] != ""

    @property
    def directory_index_specified(self) -> bool:
        """Directory Index specified conditional."""
        return self.variables["RewriteDirectoryIndex"] != ""

    @property
    def role_boundary_specified(self) -> bool:
        """IAM Role Boundary specified conditional."""
        return self.variables["RoleBoundaryArn"] != ""

    @property
    def waf_name_specified(self) -> bool:
        """WAF name specified conditional."""
        return self.variables["WAFWebACL"] != ""

    def create_template(self) -> None:
        """Create template (main function called by CFNgin)."""
        self.template.set_version("2010-09-09")
        self.template.set_description("Static Website - Bucket and Distribution")

        # Resources
        bucket = self.add_bucket()

        if self.cf_enabled:
            oai = self.add_origin_access_identity()
            bucket_policy = self.add_cloudfront_bucket_policy(bucket, oai)
            lambda_function_associations = self.get_lambda_associations()

            if self.directory_index_specified:
                index_rewrite = self._get_index_rewrite_role_function_and_version()
                lambda_function_associations = self.get_directory_index_lambda_association(
                    lambda_function_associations, index_rewrite["version"]
                )

            distribution_options = self.get_cloudfront_distribution_options(
                bucket, oai, lambda_function_associations
            )
            self.add_cloudfront_distribution(bucket_policy, distribution_options)
        else:
            self.add_bucket_policy(bucket)

    def get_lambda_associations(self) -> list[cloudfront.LambdaFunctionAssociation]:
        """Retrieve any lambda associations from the instance variables."""
        # If custom associations defined, use them
        if self.variables["lambda_function_associations"]:
            return [
                cloudfront.LambdaFunctionAssociation(
                    EventType=x["type"], LambdaFunctionARN=x["arn"]
                )
                for x in self.variables["lambda_function_associations"]
            ]
        return []

    @staticmethod
    def get_directory_index_lambda_association(
        lambda_associations: list[cloudfront.LambdaFunctionAssociation],
        directory_index_rewrite_version: awslambda.Version,
    ) -> list[cloudfront.LambdaFunctionAssociation]:
        """Retrieve the directory index lambda associations with the added rewriter.

        Args:
            lambda_associations: The lambda associations.
            directory_index_rewrite_version: The directory index rewrite version.

        """
        lambda_associations.append(
            cloudfront.LambdaFunctionAssociation(
                EventType="origin-request",
                LambdaFunctionARN=directory_index_rewrite_version.ref(),
            )
        )
        return lambda_associations

    def get_cloudfront_distribution_options(
        self,
        bucket: s3.Bucket,
        oai: cloudfront.CloudFrontOriginAccessIdentity,
        lambda_function_associations: list[cloudfront.LambdaFunctionAssociation],
    ) -> dict[str, Any]:
        """Retrieve the options for our CloudFront distribution.

        Args:
            bucket: The bucket resource
            oai: The origin access identity resource.
            lambda_function_associations: List of Lambda Function associations.

        Return:
            The CloudFront Distribution Options.

        """
        if os.getenv("AWS_REGION") == "us-east-1":
            # use global endpoint for us-east-1
            origin = Join(".", [bucket.ref(), "s3.amazonaws.com"])
        else:
            # use reginal endpoint to avoid "temporary" redirect that can last over an hour
            # https://forums.aws.amazon.com/message.jspa?messageID=677452
            origin = Join(".", [bucket.ref(), "s3", Region, "amazonaws.com"])

        return {
            "Aliases": self.add_aliases(),
            "Origins": [
                cloudfront.Origin(
                    DomainName=origin,
                    S3OriginConfig=cloudfront.S3OriginConfig(
                        OriginAccessIdentity=Join(
                            "", ["origin-access-identity/cloudfront/", oai.ref()]
                        )
                    ),
                    Id="S3Origin",
                )
            ],
            "DefaultCacheBehavior": cloudfront.DefaultCacheBehavior(
                AllowedMethods=["GET", "HEAD"],
                Compress=self.variables.get("Compress", True),
                DefaultTTL="86400",
                ForwardedValues=cloudfront.ForwardedValues(
                    Cookies=cloudfront.Cookies(Forward="none"), QueryString=False
                ),
                LambdaFunctionAssociations=lambda_function_associations,
                TargetOriginId="S3Origin",
                ViewerProtocolPolicy="redirect-to-https",
            ),
            "DefaultRootObject": "index.html",
            "Logging": self.add_logging_bucket(),
            "PriceClass": self.variables["PriceClass"],
            "CustomErrorResponses": [
                cloudfront.CustomErrorResponse(
                    ErrorCode=response["ErrorCode"],
                    ResponseCode=response["ResponseCode"],
                    ResponsePagePath=response["ResponsePagePath"],
                )
                for response in self.variables["custom_error_responses"]
            ],
            "Enabled": True,
            "WebACLId": self.add_web_acl(),
            "ViewerCertificate": self.add_acm_cert(),
        }

    def add_aliases(self) -> list[str] | Ref:
        """Add aliases."""
        if self.aliases_specified:
            return self.variables["Aliases"]
        return NoValue

    def add_web_acl(self) -> list[str] | Ref:
        """Add Web ACL."""
        if self.waf_name_specified:
            return self.variables["WAFWebACL"]
        return NoValue

    def add_logging_bucket(self) -> cloudfront.Logging | Ref:
        """Add Logging Bucket."""
        if self.cf_logging_enabled:
            return cloudfront.Logging(
                Bucket=Join(".", [self.variables["LogBucketName"], "s3.amazonaws.com"])
            )
        return NoValue

    def add_acm_cert(self) -> cloudfront.ViewerCertificate | Ref:
        """Add ACM cert."""
        if self.acm_certificate_specified:
            return cloudfront.ViewerCertificate(
                AcmCertificateArn=self.variables["AcmCertificateArn"],
                SslSupportMethod="sni-only",
            )
        return NoValue

    def add_origin_access_identity(self) -> cloudfront.CloudFrontOriginAccessIdentity:
        """Add the origin access identity resource to the template."""
        return self.template.add_resource(
            cloudfront.CloudFrontOriginAccessIdentity(
                "OAI",
                CloudFrontOriginAccessIdentityConfig=cloudfront.CloudFrontOriginAccessIdentityConfig(
                    Comment="CF access to website"
                ),
            )
        )

    def add_bucket_policy(self, bucket: s3.Bucket) -> s3.BucketPolicy:
        """Add a policy to the bucket if CloudFront is disabled. Ensure PublicRead.

        Args:
            bucket: The bucket resource to place the policy.

        Returns:
            The Bucket Policy Resource.

        """
        return self.template.add_resource(
            s3.BucketPolicy(
                "BucketPolicy",
                Bucket=bucket.ref(),
                PolicyDocument=Policy(
                    Version="2012-10-17",
                    Statement=[
                        Statement(
                            Effect=Allow,
                            Principal=Principal("*"),
                            Action=[Action("s3", "getObject")],
                            Resource=[Join("", [bucket.get_att("Arn"), "/*"])],
                        )
                    ],
                ),
            )
        )

    def add_bucket(self) -> s3.Bucket:
        """Add the bucket resource along with an output of it's name / website url.

        Returns:
            The bucket resource.

        """
        bucket = self.template.add_resource(
            s3.Bucket(
                "Bucket",
                # PublicAccessBlockConfiguration=s3.PublicAccessBlockConfiguration(
                # )
                PublicAccessBlockConfiguration=(
                    s3.PublicAccessBlockConfiguration(BlockPublicAcls="true")
                    if self.cf_enabled
                    else s3.PublicAccessBlockConfiguration(BlockPublicAcls="false")
                ),
                OwnershipControls=s3.OwnershipControls(
                    Rules=[s3.OwnershipControlsRule(ObjectOwnership="ObjectWriter")]
                ),
                LifecycleConfiguration=s3.LifecycleConfiguration(
                    Rules=[s3.LifecycleRule(NoncurrentVersionExpirationInDays=90, Status="Enabled")]
                ),
                VersioningConfiguration=s3.VersioningConfiguration(Status="Enabled"),
            )
        )
        self.template.add_output(
            Output("BucketName", Description="Name of website bucket", Value=bucket.ref())
        )

        if not self.cf_enabled:
            # bucket cannot be configured with WebsiteConfiguration when using OAI S3Origin
            bucket.WebsiteConfiguration = s3.WebsiteConfiguration(
                IndexDocument="index.html", ErrorDocument="error.html"
            )
            self.template.add_output(
                Output(
                    "BucketWebsiteURL",
                    Description="URL of the bucket website",
                    Value=bucket.get_att("WebsiteURL"),
                )
            )

        return bucket

    def add_cloudfront_bucket_policy(
        self, bucket: s3.Bucket, oai: cloudfront.CloudFrontOriginAccessIdentity
    ) -> s3.BucketPolicy:
        """Given a bucket and oai resource add cloudfront access to the bucket.

        Keyword Args:
            bucket: A bucket resource.
            oai: An Origin Access Identity resource.

        Return:
            The CloudFront Bucket access resource.

        """
        return self.template.add_resource(
            s3.BucketPolicy(
                "AllowCFAccess",
                Bucket=bucket.ref(),
                PolicyDocument=PolicyDocument(
                    Version="2012-10-17",
                    Statement=self._get_cloudfront_bucket_policy_statements(bucket, oai),
                ),
            )
        )

    def add_lambda_execution_role(
        self, name: str = "LambdaExecutionRole", function_name: str = ""
    ) -> iam.Role:
        """Create the Lambda@Edge execution role.

        Args:
            name: Name for the Lambda Execution Role.
            function_name: Name of the Lambda Function the Role will be
                attached to.

        """
        lambda_resource = Join(
            "",
            [
                "arn:",
                Partition,
                ":logs:*:",
                AccountId,
                ":log-group:/aws/lambda/",
                StackName,
                f"-{function_name}-*",
            ],
        )

        edge_resource = Join(
            "",
            [
                "arn:",
                Partition,
                ":logs:*:",
                AccountId,
                ":log-group:/aws/lambda/*.",
                StackName,
                f"-{function_name}-*",
            ],
        )

        return self.template.add_resource(
            iam.Role(
                name,
                AssumeRolePolicyDocument=make_simple_assume_policy(
                    "lambda.amazonaws.com", "edgelambda.amazonaws.com"
                ),
                PermissionsBoundary=(
                    self.variables["RoleBoundaryArn"] if self.role_boundary_specified else NoValue
                ),
                Policies=[
                    iam.Policy(
                        PolicyName="LambdaLogCreation",
                        PolicyDocument=PolicyDocument(
                            Version="2012-10-17",
                            Statement=[
                                Statement(
                                    Action=[
                                        awacs.logs.CreateLogGroup,
                                        awacs.logs.CreateLogStream,
                                        awacs.logs.PutLogEvents,
                                    ],
                                    Effect=Allow,
                                    Resource=[lambda_resource, edge_resource],
                                )
                            ],
                        ),
                    ),
                ],
            )
        )

    def add_cloudfront_directory_index_rewrite(self, role: iam.Role) -> awslambda.Function:
        """Add an index CloudFront directory index rewrite lambda function to the template.

        Keyword Args:
            role: The index rewrite role resource.

        Return:
            The CloudFront directory index rewrite lambda function resource.

        """
        code_str = ""
        path = os.path.join(  # noqa: PTH118
            os.path.dirname(__file__),  # noqa: PTH120
            "templates/cf_directory_index_rewrite.template.js",
        )
        with open(path, encoding="utf-8") as file_:  # noqa: PTH123
            code_str = file_.read().replace(
                "{{RewriteDirectoryIndex}}", self.variables["RewriteDirectoryIndex"]
            )

        function = self.template.add_resource(
            awslambda.Function(
                "CFDirectoryIndexRewrite",
                Code=awslambda.Code(ZipFile=code_str),
                DeletionPolicy="Retain",
                Description="Rewrites CF directory HTTP requests to default page",
                Handler="index.handler",
                Role=role.get_att("Arn"),
                Runtime="nodejs10.x",
            )
        )

        self.template.add_output(
            Output(
                "LambdaCFDirectoryIndexRewriteArn",
                Description="Directory Index Rewrite Function Arn",
                Value=function.get_att("Arn"),
            )
        )

        return function

    def add_cloudfront_directory_index_rewrite_version(
        self, directory_index_rewrite: awslambda.Function
    ) -> awslambda.Version:
        """Add a specific version to the directory index rewrite lambda.

        Args:
            directory_index_rewrite: The directory index rewrite lambda resource.

        Return:
            The CloudFront directory index rewrite version.

        """
        code_hash = hashlib.md5(  # noqa: S324
            str(directory_index_rewrite.properties["Code"].properties["ZipFile"]).encode()
        ).hexdigest()

        return self.template.add_resource(
            awslambda.Version(
                "CFDirectoryIndexRewriteVer" + code_hash,
                FunctionName=directory_index_rewrite.ref(),
            )
        )

    def add_cloudfront_distribution(
        self,
        bucket_policy: s3.BucketPolicy,
        cloudfront_distribution_options: dict[str, Any],
    ) -> cloudfront.Distribution:
        """Add the CloudFront distribution to the template / output the id and domain name.

        Args:
            bucket_policy: Bucket policy to allow CloudFront access.
            cloudfront_distribution_options: The distribution options.

        Return:
            The CloudFront Distribution resource

        """
        distribution = self.template.add_resource(
            cloudfront.Distribution(
                "CFDistribution",
                DependsOn=bucket_policy.title,
                DistributionConfig=cloudfront.DistributionConfig(**cloudfront_distribution_options),
            )
        )
        self.template.add_output(
            Output(
                "CFDistributionId",
                Description="CloudFront distribution ID",
                Value=distribution.ref(),
            )
        )
        self.template.add_output(
            Output(
                "CFDistributionDomainName",
                Description="CloudFront distribution domain name",
                Value=distribution.get_att("DomainName"),
            )
        )
        return distribution

    @staticmethod
    def _get_cloudfront_bucket_policy_statements(
        bucket: s3.Bucket, oai: cloudfront.CloudFrontOriginAccessIdentity
    ) -> list[Statement]:
        return [
            Statement(
                Action=[awacs.s3.GetObject],
                Effect=Allow,
                # S3CanonicalUserId is translated to the ARN when AWS renders this
                Principal=Principal("CanonicalUser", oai.get_att("S3CanonicalUserId")),
                Resource=[Join("", [bucket.get_att("Arn"), "/*"])],
            )
        ]

    def _get_index_rewrite_role_function_and_version(
        self,
    ) -> _IndexRewriteFunctionInfoTypeDef:
        role = self.add_lambda_execution_role(
            "CFDirectoryIndexRewriteRole", "CFDirectoryIndexRewrite"
        )
        function = self.add_cloudfront_directory_index_rewrite(role)
        version = self.add_cloudfront_directory_index_rewrite_version(function)
        return _IndexRewriteFunctionInfoTypeDef(function=function, role=role, version=version)


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    print(  # noqa: T201
        StaticSite("test", CfnginContext(parameters={"namespace": "test"})).to_json()
    )



"""Base module for test handlers."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ...config.components.runway.base import ConfigProperty


class TestHandler:
    """Base class for test handlers."""

    @classmethod
    def handle(cls, name: str, args: ConfigProperty | dict[str, Any]) -> None:
        """Redefine in subclass."""
        raise NotImplementedError

    @staticmethod
    def get_dirs(provided_path: str) -> list[str]:
        """Return list of directories."""
        repo_dirs = next(os.walk(provided_path))[1]
        if ".git" in repo_dirs:
            repo_dirs.remove(".git")  # not relevant for any repo operations
        return repo_dirs



"""Runway ``future`` definition model."""

from pydantic import ConfigDict

from ..base import ConfigProperty


class RunwayFutureDefinitionModel(ConfigProperty):
    """Enable features/behaviors that will be become standard ahead of their official release."""

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "description": "Enable features/behaviors that will be become standard "
            "ahead of their official release."
        },
        title="Runway Future Definition",
        validate_default=True,
        validate_assignment=True,
    )



"""``runway gen-sample cfn` command."""

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import click

from ... import options
from .utils import TEMPLATES, copy_sample, write_tfstate_template

if TYPE_CHECKING:
    from ...._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("cfn", short_help="cfngin + cfn (sampleapp.cfn)")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def cfn(ctx: click.Context, **_: Any) -> None:
    """Generate a sample CFNgin project using CloudFormation."""
    src = TEMPLATES / "cfn"
    dest = Path.cwd() / "sampleapp.cfn"
    templates = dest / "templates"
    tf_state = templates / "tf_state.yml"

    copy_sample(ctx, src, dest)
    templates.mkdir()
    write_tfstate_template(tf_state)
    LOGGER.success("Sample CloudFormation module created at %s", dest)



"""Execute the AWS CLI update-kubeconfig command."""

from __future__ import annotations

import logging
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any

from runway.config import RunwayConfig

if TYPE_CHECKING:
    from runway.context import CfnginContext

LOGGER = logging.getLogger(__name__)


def copy_template_to_env(path: Path, env: str, region: str) -> None:
    """Copy k8s module template into new environment directory."""
    overlays_dir = path / "overlays"
    template_dir = overlays_dir / "template"
    env_dir = overlays_dir / env
    if template_dir.is_dir():
        if env_dir.is_dir() or (Path(f"{env_dir}-{region}").is_dir()):
            LOGGER.info(
                'Bootstrap of k8s module at "%s" skipped; module '
                "already has a config for this environment",
                path,
            )
        else:
            LOGGER.info(
                'Copying overlay template at "%s" to new environment directory "%s"',
                template_dir,
                env_dir,
            )
            shutil.copytree(template_dir, env_dir, symlinks=True)
            # Update templated environment name in files
            for i in [
                "kustomization.yaml",
                # namespace files can't be directly kustomized
                "namespace.yaml",
            ]:
                templated_file_path = env_dir / i
                if templated_file_path.is_file():
                    filedata = templated_file_path.read_text()
                    if "REPLACE_ME_ENV" in filedata:
                        templated_file_path.write_text(filedata.replace("REPLACE_ME_ENV", env))
    else:
        LOGGER.info(
            'Skipping bootstrap of k8s module at "%s"; no template directory present',
            path,
        )


def create_runway_environments(*, context: CfnginContext, namespace: str, **_: Any) -> bool:
    """Copy k8s module templates into new environment directories.

    Args:
        context: CFNgin context.
        namespace: Current CFNgin namespace.

    Returns: boolean for whether or not the hook succeeded.

    """
    LOGGER.info("Bootstrapping runway k8s modules, looking for unconfigured environments...")

    environment = namespace

    runway_config_path = RunwayConfig.find_config_file(Path.cwd().parent)
    if not runway_config_path.is_file():
        LOGGER.warning("could not find RUNWAYCONFIG=%s", runway_config_path)
        return False

    runway_config = RunwayConfig.parse_file(file_path=runway_config_path)

    for deployment in runway_config.deployments:
        for module in deployment.modules:
            if isinstance(module.path, Path):
                path = module.path.name
            elif isinstance(module.path, str):
                path = module.path
            else:
                path = ""
            if path.endswith(".k8s"):
                copy_template_to_env(
                    runway_config_path.parent / path,
                    environment,
                    context.env.aws_region,
                )
    return True



"""Docker image push hook.

Replicates the functionality of the ``docker image push`` CLI command.

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Annotated, Any

from pydantic import ConfigDict, Field, field_validator

from .....context import CfnginContext
from .....utils import BaseModel
from ..data_models import (
    DockerImage,
    ElasticContainerRegistry,
    ElasticContainerRegistryRepository,
)
from ..hook_data import DockerHookData

if TYPE_CHECKING:
    from pydantic import ValidationInfo

LOGGER = logging.getLogger(__name__.replace("._", "."))


class ImagePushArgs(BaseModel):
    """Args passed to image.push."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    ctx: Annotated[CfnginContext | None, Field(alias="context", exclude=True)] = None

    ecr_repo: ElasticContainerRegistryRepository | None = None  # depends on ctx
    """AWS Elastic Container Registry repository information.
    Providing this will automatically construct the repo URI.
    If provided, do not provide ``repo``.

    If using a private registry, only ``repo_name`` is required.
    If using a public registry, ``repo_name`` and ``registry_alias``.

    """

    image: DockerImage | None = None
    """Image to push."""

    repo: Annotated[str | None, Field(validate_default=True)] = None  # depends on ecr_repo & image
    """URI of a non Docker Hub repository where the image will be stored."""

    tags: Annotated[list[str], Field(validate_default=True)] = []  # depends on image
    """List of tags to push."""

    @field_validator("ecr_repo", mode="before")
    @classmethod
    def _set_ecr_repo(cls, v: Any, info: ValidationInfo) -> Any:
        """Set the value of ``ecr_repo``."""
        if v and isinstance(v, dict):
            return ElasticContainerRegistryRepository.model_validate(
                {
                    "repo_name": v.get("repo_name"),
                    "registry": ElasticContainerRegistry.model_validate(
                        {
                            "account_id": v.get("account_id"),
                            "alias": v.get("registry_alias"),
                            "aws_region": v.get("aws_region"),
                            "context": info.data.get("context"),
                        }
                    ),
                }
            )
        return v

    @field_validator("repo", mode="before")
    @classmethod
    def _set_repo(cls, v: str | None, info: ValidationInfo) -> str | None:
        """Set the value of ``repo``."""
        if v:
            return v

        image: DockerImage | None = info.data.get("image")
        if image:
            return image.repo

        ecr_repo: ElasticContainerRegistryRepository | None = info.data.get("ecr_repo")
        if ecr_repo:
            return ecr_repo.fqn

        return None

    @field_validator("tags", mode="before")
    @classmethod
    def _set_tags(cls, v: list[str], info: ValidationInfo) -> list[str]:
        """Set the value of ``tags``."""
        if v:
            return v

        image: DockerImage | None = info.data.get("image")
        if image:
            return image.tags

        return ["latest"]


def push(*, context: CfnginContext, **kwargs: Any) -> DockerHookData:
    """Docker image push hook.

    Replicates the functionality of ``docker image push`` CLI command.

    kwargs are parsed by :class:`~runway.cfngin.hooks.docker.image.ImagePushArgs`.

    """
    args = ImagePushArgs.model_validate({"context": context, **kwargs})
    docker_hook_data = DockerHookData.from_cfngin_context(context)
    LOGGER.info("pushing image %s...", args.repo)
    for tag in args.tags:
        docker_hook_data.client.images.push(repository=args.repo, tag=tag)
        LOGGER.info("successfully pushed image %s:%s", args.repo, tag)
    return docker_hook_data.update_context(context)



"""``runway gen-sample static-angular`` command."""

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import click

from ... import options
from .utils import TEMPLATES, convert_gitignore, copy_sample

if TYPE_CHECKING:
    from ...._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("static-angular", short_help="angular static site (static-angular)")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def static_angular(ctx: click.Context, **_: Any) -> None:
    """Generate a sample static site project using Angular."""
    src = TEMPLATES / "static-angular"
    dest = Path.cwd() / "static-angular"

    copy_sample(ctx, src, dest)
    convert_gitignore(dest / "sampleapp.web/_gitignore")

    LOGGER.success("Sample static Angular site repo created at %s", dest)
    LOGGER.notice("See the README for setup and deployment instructions.")



"""``runway docs`` command."""

# docs: file://./../../../docs/source/commands.rst
import logging
import os
from typing import TYPE_CHECKING, Any, cast

import click

from ...utils import SafeHaven
from .. import options

if TYPE_CHECKING:
    from ..._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))

DOCS_URL = "https://runway.readthedocs.io/"


@click.command("docs", short_help="open doc site")
@options.debug
@options.no_color
@options.verbose
def docs(**_: Any) -> None:
    """Open the Runway documentation web site using the default web browser."""
    with SafeHaven():
        # Pyinstaller sets this var on systems to force using the internal lib
        # but, this can break some functionally around opening a web browser
        # using click so, reset/remove the var.
        lp_key = "LD_LIBRARY_PATH"  # for GNU/Linux and *BSD.
        lp_orig = os.getenv(lp_key + "_ORIG")
        if lp_orig:
            LOGGER.debug("temporarily reverting environ: %s=%s", lp_key, lp_orig)
            os.environ[lp_key] = lp_orig  # restore the original, unmodified value
        else:
            # This happens when LD_LIBRARY_PATH was not set.
            # Remove the env var as a last resort:
            LOGGER.debug("temporarily removing environ: %s", lp_key)
            os.environ.pop(lp_key, None)
        LOGGER.verbose("launching url: %s", DOCS_URL)
        click.launch(DOCS_URL)



"""Python dependency compatibility handling."""

import sys
from functools import cached_property
from importlib.metadata import PackageNotFoundError, version
from shlex import join as shlex_join

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

__all__ = [
    "PackageNotFoundError",  # TODO (kyle): remove in next major release
    "Self",
    "cached_property",  # TODO (kyle): remove in next major release
    "shlex_join",  # TODO (kyle): remove in next major release
    "version",  # TODO (kyle): remove in next major release
]



"""Exceptions for awslambda hooks."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ...exceptions import CfnginError

if TYPE_CHECKING:
    from pathlib import Path


class DeploymentPackageEmptyError(CfnginError):
    """Deployment package is empty.

    This can be caused by an incorrect source code directory or a gitignore rule
    unintentionally ignoring all source code.

    Any empty deployment package is determined by checking the size of the
    archive file. If the size is <=22 (the size a zip file End of Central
    Directory Record) it has no contents.

    """

    archive_file: Path
    """The deployment package archive file."""

    def __init__(self, archive_file: Path) -> None:
        """Instantiate class.

        Args:
            archive_file: The empty archive file.

        """
        self.archive_file = archive_file
        self.message = f"{archive_file.name} contains no files"
        super().__init__()


class RuntimeMismatchError(CfnginError):
    """Required runtime does not match the detected runtime."""

    detected_runtime: str
    """Runtime detected on the build system."""

    expected_runtime: str
    """Explicitly defined runtime that was expected."""

    def __init__(self, expected_runtime: str, detected_runtime: str) -> None:
        """Instantiate class.

        Args:
            expected_runtime: Explicitly defined runtime that was expected.
            detected_runtime: Runtime detected on the build system.

        """
        self.detected_runtime = detected_runtime
        self.expected_runtime = expected_runtime
        self.message = (
            f"{detected_runtime} runtime determined from the build system"
            f" does not match the expected {expected_runtime} runtime"
        )
        super().__init__()



#!/usr/bin/env python
"""Module with static website supporting infrastructure."""

import logging
from typing import Any

import awacs.s3
from awacs.aws import Allow, AWSPrincipal, Policy, Statement
from troposphere import AccountId, Join, Output, cognito, s3

from ...cfngin.blueprints.base import Blueprint
from ...cfngin.hooks.staticsite.auth_at_edge.client_updater import get_redirect_uris
from ...module.staticsite.utils import add_url_scheme

LOGGER = logging.getLogger(__name__)


class Dependencies(Blueprint):
    """CFNgin blueprint for creating static website buckets."""

    VARIABLES = {
        "Aliases": {
            "type": list,
            "default": [],
            "description": "(Optional) Domain aliases for the distribution",
        },
        "AdditionalRedirectDomains": {
            "type": list,
            "default": [],
            "description": "(Optional) AppClient callback/logout domains (in "
            "addition to the Aliases)",
        },
        "AuthAtEdge": {
            "type": bool,
            "default": False,
            "description": "Utilizing Authorization @ Edge",
        },
        "CreateUserPool": {
            "type": bool,
            "default": False,
            "description": "Whether a User Pool should be created for the project",
        },
        "SupportedIdentityProviders": {
            "type": list,
            "default": [],
            "description": "Auth@Edge AppClient Identity Providers",
        },
        "OAuthScopes": {
            "type": list,
            "default": [
                "phone",
                "email",
                "profile",
                "openid",
                "aws.cognito.signin.user.admin",
            ],
            "description": "The allowed scopes for OAuth validation",
        },
        "RedirectPathSignIn": {
            "type": str,
            "default": "/parseauth",
            "description": "Auth@Edge redirect sign in path",
        },
        "RedirectPathSignOut": {
            "type": str,
            "default": "/",
            "description": "Auth@Edge redirect sign out path",
        },
    }

    def create_template(self) -> None:
        """Create template (main function called by CFNgin)."""
        template = self.template
        template.set_version("2010-09-09")
        template.set_description("Static Website - Dependencies")

        # Resources
        awslogbucket = template.add_resource(
            s3.Bucket(
                "AWSLogBucket",
                AccessControl=s3.Private,
                VersioningConfiguration=s3.VersioningConfiguration(Status="Enabled"),
            )
        )
        template.add_output(
            Output(
                "AWSLogBucketName",
                Description="Name of bucket storing AWS logs",
                Value=awslogbucket.ref(),
            )
        )

        template.add_resource(
            s3.BucketPolicy(
                "AllowAWSLogWriting",
                Bucket=awslogbucket.ref(),
                PolicyDocument=Policy(
                    Version="2012-10-17",
                    Statement=[
                        Statement(
                            Action=[awacs.s3.PutObject],
                            Effect=Allow,
                            Principal=AWSPrincipal(Join(":", ["arn:aws:iam:", AccountId, "root"])),
                            Resource=[Join("", ["arn:aws:s3:::", awslogbucket.ref(), "/*"])],
                        )
                    ],
                ),
            )
        )
        artifacts = template.add_resource(
            s3.Bucket(
                "Artifacts",
                AccessControl=s3.Private,
                LifecycleConfiguration=s3.LifecycleConfiguration(
                    Rules=[s3.LifecycleRule(NoncurrentVersionExpirationInDays=90, Status="Enabled")]
                ),
                VersioningConfiguration=s3.VersioningConfiguration(Status="Enabled"),
            )
        )
        template.add_output(
            Output(
                "ArtifactsBucketName",
                Description="Name of bucket storing artifacts",
                Value=artifacts.ref(),
            )
        )

        if self.variables["AuthAtEdge"]:
            userpool_client_params: dict[str, Any] = {
                "AllowedOAuthFlows": ["code"],
                "AllowedOAuthScopes": self.variables["OAuthScopes"],
            }
            if self.variables["Aliases"]:
                userpool_client_params["AllowedOAuthFlowsUserPoolClient"] = True
                userpool_client_params["SupportedIdentityProviders"] = self.variables[
                    "SupportedIdentityProviders"
                ]

                redirect_domains = [add_url_scheme(x) for x in self.variables["Aliases"]] + [
                    add_url_scheme(x) for x in self.variables["AdditionalRedirectDomains"]
                ]
                redirect_uris = get_redirect_uris(
                    redirect_domains,
                    self.variables["RedirectPathSignIn"],
                    self.variables["RedirectPathSignOut"],
                )
                userpool_client_params["CallbackURLs"] = redirect_uris["sign_in"]
                userpool_client_params["LogoutURLs"] = redirect_uris["sign_out"]
            else:
                userpool_client_params["CallbackURLs"] = self.context.hook_data[
                    "aae_callback_url_retriever"
                ]["callback_urls"]

            if self.variables["CreateUserPool"]:
                user_pool = template.add_resource(cognito.UserPool("AuthAtEdgeUserPool"))

                user_pool_id = user_pool.ref()

                template.add_output(
                    Output(
                        "AuthAtEdgeUserPoolId",
                        Description="Cognito User Pool App Client for Auth @ Edge",
                        Value=user_pool_id,
                    )
                )
            else:
                user_pool_id = self.context.hook_data["aae_user_pool_id_retriever"]["id"]
            userpool_client_params["UserPoolId"] = user_pool_id

            client = template.add_resource(
                cognito.UserPoolClient("AuthAtEdgeClient", **userpool_client_params)
            )

            template.add_output(
                Output(
                    "AuthAtEdgeClient",
                    Description="Cognito User Pool App Client for Auth @ Edge",
                    Value=client.ref(),
                )
            )


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    from runway.context import CfnginContext

    print(  # noqa: T201
        Dependencies("test", CfnginContext(parameters={"namespace": "test"})).to_json()
    )



"""User Pool Client Domain Updater."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from ...base import HookArgsBaseModel

if TYPE_CHECKING:
    from .....context import CfnginContext

LOGGER = logging.getLogger(__name__)


class HookArgs(HookArgsBaseModel):
    """Hook arguments."""

    client_id: str
    """The ID of the Cognito User Pool Client."""


def update(context: CfnginContext, *_args: Any, **kwargs: Any) -> dict[str, Any] | bool:
    """Retrieve/Update the domain name of the specified client.

    A domain name is required in order to make authorization and token
    requests. This prehook ensures we have one available, and if not
    we create one based on the user pool and client ids.

    Arguments parsed by
    :class:`~runway.cfngin.hooks.staticsite.auth_at_edge.domain_updater.HookArgs`.

    Args:
        context: The context instance.
        **kwargs: Arbitrary keyword arguments.

    """
    args = HookArgs.model_validate(kwargs)
    session = context.get_session()
    cognito_client = session.client("cognito-idp")

    context_dict: dict[str, Any] = {}

    user_pool_id = context.hook_data["aae_user_pool_id_retriever"]["id"]
    user_pool = cognito_client.describe_user_pool(UserPoolId=user_pool_id).get("UserPool", {})
    (user_pool_region, user_pool_hash) = user_pool_id.split("_")

    domain_prefix = user_pool.get("CustomDomain", user_pool.get("Domain"))

    # Return early if we already have a domain
    if domain_prefix:
        context_dict["domain"] = get_user_pool_domain(domain_prefix, user_pool_region)
        return context_dict

    try:
        domain_prefix = (f"{user_pool_hash}-{args.client_id}").lower()

        cognito_client.create_user_pool_domain(Domain=domain_prefix, UserPoolId=user_pool_id)
        context_dict["domain"] = get_user_pool_domain(domain_prefix, user_pool_region)
        return context_dict
    except Exception:
        LOGGER.exception("could not update user pool domain: %s", user_pool_id)
        return False


def delete(context: CfnginContext, *_args: Any, **kwargs: Any) -> dict[str, Any] | bool:
    """Delete the domain if the user pool was created by Runway.

    If a User Pool was created by Runway, and populated with a domain, that
    domain must be deleted prior to the User Pool itself being deleted or an
    error will occur. This process ensures that our generated domain name is
    deleted, or skips if not able to find one.

    Arguments parsed by
    :class:`~runway.cfngin.hooks.staticsite.auth_at_edge.domain_updater.HookArgs`.

    Args:
        context: The context instance.
        **kwargs: Arbitrary keyword arguments.

    """
    args = HookArgs.model_validate(kwargs)
    session = context.get_session()
    cognito_client = session.client("cognito-idp")

    user_pool_id = context.hook_data["aae_user_pool_id_retriever"]["id"]
    _, user_pool_hash = user_pool_id.split("_")
    domain_prefix = (f"{user_pool_hash}-{args.client_id}").lower()

    try:
        cognito_client.delete_user_pool_domain(UserPoolId=user_pool_id, Domain=domain_prefix)
        return True
    except cognito_client.exceptions.InvalidParameterException:
        LOGGER.info('skipped deletion; no domain with prefix "%s"', domain_prefix)
        return True
    except Exception:
        LOGGER.exception("could not delete user pool domain")
        return False


def get_user_pool_domain(prefix: str, region: str) -> str:
    """Return a user pool domain name based on the prefix received and region.

    Args:
        prefix: The domain prefix for the domain.
        region: The region in which the pool resides.

    """
    return f"{prefix}.auth.{region}.amazoncognito.com"



"""Base models & other objects."""

from __future__ import annotations

from pydantic import ConfigDict

from ...utils import BaseModel


class ConfigProperty(BaseModel):
    """Base class for Runway configuration properties."""

    model_config = ConfigDict(
        extra="ignore",
        validate_default=True,
        validate_assignment=True,
    )



"""Runway ``assume_role`` definition model."""

from __future__ import annotations

from typing import Annotated

from pydantic import ConfigDict, Field, field_validator

from .. import utils
from ..base import ConfigProperty
from ..utils import RUNWAY_LOOKUP_STRING_REGEX


class RunwayAssumeRoleDefinitionModel(ConfigProperty):
    """Used to defined a role to assume while Runway is processing each module."""

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "examples": [
                {"arn": "arn:aws:iam::123456789012:role/name"},
                {
                    "arn": "${var role_arn.${env DEPLOY_ENVIRONMENT}}",
                    "duration": 9001,
                    "post_deploy_env_revert": True,
                    "session_name": "runway-example",
                },
            ],
        },
        title="Runway Deployment.assume_role Definition",
        validate_default=True,
        validate_assignment=True,
    )

    arn: Annotated[
        str | None,
        Field(description="The ARN of the AWS IAM role to be assumed. (supports lookups)"),
    ] = None
    """The ARN of the AWS IAM role to be assumed. (supports lookups)"""

    duration: (
        Annotated[
            int,
            Field(
                description="The duration, in seconds, of the role session. (supports lookups)",
                ge=900,
                le=43200,
            ),
        ]
        | Annotated[str, Field(pattern=RUNWAY_LOOKUP_STRING_REGEX)]
    ) = 3600
    """The duration, in seconds, of the role session. (supports lookups)"""

    post_deploy_env_revert: Annotated[
        bool,
        Field(
            title="Post Deployment Environment Revert",
            description="Revert the credentials stored in environment variables to "
            "what they were prior to execution after the deployment finished processing. "
            "(supports lookups)",
        ),
    ] = False
    """Revert the credentials stored in environment variables to what they were
    prior to execution after the deployment finished processing. (supports lookups)

    """

    session_name: Annotated[
        str,
        Field(
            description="An identifier for the assumed role session. (supports lookups)",
        ),
    ] = "runway"
    """An identifier for the assumed role session. (supports lookups)"""

    @field_validator("arn")
    @classmethod
    def _convert_arn_null_value(cls, v: str | None) -> str | None:
        """Convert a "nul" string into type(None)."""
        null_strings = ["null", "none", "undefined"]
        return None if isinstance(v, str) and v.lower() in null_strings else v

    @field_validator("duration", mode="before")
    @classmethod
    def _validate_duration(cls, v: int | str) -> int | str:
        """Validate duration is within the range allowed by AWS."""
        if isinstance(v, str):
            return v
        if v < 900:
            raise ValueError("duration must be greater than or equal to 900")
        if v > 43_200:
            raise ValueError("duration must be less than or equal to 43,200")
        return v

    _validate_string_is_lookup = field_validator("duration", mode="before")(
        utils.validate_string_is_lookup
    )
